/*
 * tclFCmd.c
 *
 *      This file implements the generic portion of file manipulation 
 *      subcommands of the "file" command. 
 *
 * Copyright (c) 1996 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tclFCmd.c 1.5 96/08/28 16:48:22
 */

#include "tclInt.h"
#include "tclPort.h"

/*
 * Declarations for local procedures defined in this file:
 */

static int        renamefile _ANSI_ARGS_((Tcl_Interp *interp, char *source, 
		      char *dest, int force)) ;
static int        copyfile _ANSI_ARGS_((Tcl_Interp *interp, char *source, 
		      char *dest, int force)) ;


/*
 *----------------------------------------------------------------------
 *
 * TclRenameFilesCmd
 *
 *	This procedure implements the "rename" subcommand of the "file"
 *      command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

    /* ARGSUSED */
int
TclRenameFilesCmd(interp, argc, argv)
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    int i = 0;                          /* loop counter */
    struct stat sb;                  /* info returned by stat() command */
    Tcl_DString buffer;                    
    Tcl_DString targetBuffer ;
    char *filename ;
    char *targetFileName ;
    int pargc ;
    char **pargv = NULL;
    int r ;                        /* return value of Move function */
    char **jargv ;
    int force = 0;

    while (i < argc) {
	if (*argv[i] == '-') {
	    if (strcmp(argv[i], "-force") == 0) {
		force = 1;
	    } else if (strcmp(argv[i], "--") == 0) {
		argc-- ;
		argv++ ;
		break ;
	    } else {
		Tcl_AppendResult(interp, "bad option \"", argv[i], 
				 " should be -force or --", (char *)NULL) ;
		return TCL_ERROR ;
	    }
	    argc-- ;
	    argv++ ;
	} else {
	    break ;
	}
    }

    /* 
     * The first element of argv should be the first argument after
     * the "rename" subcommand.
     */
    if (argc < 2) {
	Tcl_AppendResult(interp, 
			 "wrong # of args: should be \n", 
			 "\"file rename ?options? source target\" or \n", 
			 "\"file rename ?options? source ?source ... ? targetdir\"", 
			 (char *) NULL) ;
		return TCL_ERROR ;
    }	
    

    /*
     * If the stat on the target fails or the target isn't a directory,
     * try the move.  More than 2 arguments is an error in this case.
     */
    Tcl_DStringInit(&targetBuffer) ;
    targetFileName = Tcl_TranslateFileName(interp, argv[argc - 1], 
					   &targetBuffer) ;
    if (targetFileName == NULL) {
	Tcl_AppendResult(interp, "illegal file name \"", 
			 argv[argc-1], "\"", (char *)NULL) ;
	Tcl_DStringFree(&targetBuffer) ;
	return TCL_ERROR;
    }

    if (stat(targetFileName, &sb) || !S_ISDIR(sb.st_mode)) {
	if (argc > 2) {
	    Tcl_AppendResult(interp, "target \"", argv[argc - 1],
			     "\" is not a directory", (char *) NULL);
	    Tcl_DStringFree(&targetBuffer) ;
	    return TCL_ERROR;
	}
	r = renamefile(interp, argv[0], targetFileName, force) ;
	Tcl_DStringFree(&targetBuffer) ;
	
	return r ;
    }
    
    
    /*
     * If the target is a directory, then move each file into it.
     */
    
    Tcl_DStringInit(&buffer) ;

    for (i = 0; i < argc - 1; i++ ) {
	/*
	 * Extract the basename from each source, and
	 * append it to the end of the target path.
	 */
	Tcl_SplitPath(argv[i], &pargc, &pargv) ;
	jargv = (char **)ckalloc(2*sizeof(char *)) ;
	jargv[0] = targetFileName ;
	jargv[1] = pargv[pargc - 1] ;
	
	filename = Tcl_JoinPath(2, jargv, &buffer) ;
	
	ckfree((char *)pargv) ;
	ckfree((char *)jargv) ;
	
	r = renamefile(interp, argv[i], filename, force) ;

	Tcl_DStringFree(&buffer) ;
	if (r == TCL_ERROR) {
	    Tcl_DStringFree(&targetBuffer) ;
	    return TCL_ERROR;
	}

    }
    Tcl_DStringFree(&targetBuffer) ;
	
    return TCL_OK ;
}

/*
 *----------------------------------------------------------------------
 *
 * renamefile
 *
 *	rename a single file or directory
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *      If source is a nonempty directory, the contents of source will      
 *      be replicated inside a directory named by target
 *----------------------------------------------------------------------
 */

static int
renamefile(interp, source, target, force) 
    Tcl_Interp *interp ;
    char *source ;
    char *target ;
    int force ;
{
    int r ;
    Tcl_DString sourcePath, targetPath;
    char *targetName, *sourceName ;
    struct stat srcStatBuf, dstStatBuf ;

    Tcl_DStringInit(&sourcePath) ;
    sourceName = Tcl_TranslateFileName(interp, source, &sourcePath) ;
    if (sourceName == NULL) {
	Tcl_AppendResult(interp,"illegal file name \"",
			 source, "\"", (char *)NULL);
	Tcl_DStringFree(&sourcePath);
	return TCL_ERROR;
    }

    Tcl_DStringInit(&targetPath) ;
    targetName = Tcl_TranslateFileName(interp, target, &targetPath) ;
    if (targetName == NULL) {
	Tcl_AppendResult(interp,"illegal file name \"",
			 target, "\"", (char *)NULL);
	Tcl_DStringFree(&targetPath);
	return TCL_ERROR;
    }

    /*
     * Prevent renaming a file to a directory and vice versa.
     * The unix rename() command does this automatically, but it
     * is unclear whether the other platforms will fail or not.
     */
    if (stat(sourceName, &srcStatBuf) != 0) {
	goto defaulterror ;
    }
    if (stat(targetName, &dstStatBuf) == 0) {
	if (S_ISDIR(srcStatBuf.st_mode) && !S_ISDIR(dstStatBuf.st_mode)) {
	    Tcl_AppendResult(interp, "can't overwrite non-directory \"", 
			     source, "\" with directory \"", target, "\"", 
			     (char *) NULL) ;
	    goto error ;
	}
	if (!S_ISDIR(srcStatBuf.st_mode) && S_ISDIR(dstStatBuf.st_mode)) {
	    Tcl_AppendResult(interp, "can't overwrite directory \"", source, 
			     "\" with non-directory \"", target, "\"", 
			     (char *) NULL) ;
	    goto error ;
	}
    }
    
    /*
     * 	If rename succeeds, cp shall do nothing more with the
     *	current source file.  If it fails for any other reason than
     *	EXDEV, cp shall appends a diagnostic error to the result
     *      and returns an error
     */ 

    
    if (TclRenameFile(sourceName, targetName, force) == TCL_OK) {
	return (TCL_OK) ;
    }
	
    /* 
     * Rename failed
     */

    switch (errno) {
    case EINVAL:
	Tcl_AppendResult(interp, "can't rename \"", source, 
			 "\" to \"",target, "\" : ",
			 "trying to move a directory into itself or ",
			 "renaming a file to itself", 
			 (char *)NULL) ;
	goto error ;
	break ;

    case EXDEV:
	break ;

    case EEXIST:
	Tcl_AppendResult(interp, "can't overwrite file \"", target, "\"", 
			 (char *)NULL) ;
	goto error ;
    case ENOTDIR:
    case ENOENT:
    default:
    defaulterror:
	Tcl_AppendResult(interp, "error renaming \"", source, 
			 "\" to \"", target, "\" : ", (char *)NULL) ;
	Tcl_AppendResult(interp, Tcl_PosixError(interp), (char *)NULL) ;
    error:
	Tcl_DStringFree(&sourcePath) ;
	Tcl_DStringFree(&targetPath) ;
	return TCL_ERROR ;

    }

    /*
     * rename failed because the source and target are on different
     * filesystems. Now we handle that case.
     *
     */

    /*
     * 	The file hierarchy rooted in source_file shall be duplicated
     *	as a file hiearchy rooted in the destination path...
     */

    if (S_ISREG(srcStatBuf.st_mode)) {
	r = TclCopyFile(sourceName, targetName, force) ;
	if (r != TCL_ERROR) {
	    if (TclDeleteFile(sourceName) == TCL_ERROR) { 
		Tcl_AppendResult(interp, "can't remove \"", sourceName, 
				 "\" : ", Tcl_PosixError(interp), 
				 (char *)NULL) ;
		r = TCL_ERROR ;
	    }
	}
    } else {
	r = TclCopyDirectory(sourceName, targetName, force);
	if (r != TCL_ERROR) {
	    if (TclRmdir(sourceName, 1) == TCL_ERROR) {
		Tcl_AppendResult(interp, "can't remove \"", sourceName, 
				 "\" : ", Tcl_PosixError(interp), 
				 (char *)NULL) ;
		r = TCL_ERROR ;
	    }
	}
    }

    Tcl_DStringFree(&sourcePath) ;
    Tcl_DStringFree(&targetPath) ;
    return r ;

}

/*
 *----------------------------------------------------------------------
 *
 * TclCopyFilesCmd
 *
 *	This procedure is invoked to process the "copy" subcommand
 *      of the "file" command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

    /* ARGSUSED */
int
TclCopyFilesCmd(interp, argc, argv)
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    int i = 0;                          /* loop counter */
    struct stat statbuf;                  /* info returned by stat() command */
    Tcl_DString buffer;                    
    Tcl_DString targetBuffer ;
    char *filename ;
    char *targetFileName ;
    int pargc ;
    char **pargv = NULL;
    int r ;                        /* return value of Move function */
    char **jargv ;
    int force = 0;
	
    while (i < argc) {
	if (*argv[i] == '-') {
	    if (strcmp(argv[i], "-force") == 0) {
		force = 1;
	    } else if (strcmp(argv[i], "--") == 0) {
		argc-- ;
		argv++ ;
		break ;
	    } else {
		Tcl_AppendResult(interp, "bad option \"", argv[i], 
				 " should be -force or --", (char *)NULL) ;
		return TCL_ERROR ;
	    }
	    argc-- ;
	    argv++ ;
	} else {
	    break ;
	}
    }

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args: should be \n",
			 "\"file copy ?options? source target\" or \n",
			 "\"file copy ?options? source ?source ... ? targetdir\"", 
			 (char *) NULL) ;
	return TCL_ERROR ;
    }
    

    Tcl_DStringInit(&targetBuffer) ;
    targetFileName = Tcl_TranslateFileName(interp, argv[argc - 1], 
					   &targetBuffer) ;
    if (targetFileName == NULL) {
	Tcl_AppendResult(interp, "illegal file name \"", 
			 argv[argc-1], "\"", (char *)NULL) ;
	Tcl_DStringFree(&targetBuffer) ;
	return TCL_ERROR;
    }

    /*
     * If the stat on the target fails or the target isn't a directory,
     * assume the first form of the cp command. 
     * More than 2 arguments is an error in this case.
     */
    if (stat(targetFileName, &statbuf) || !S_ISDIR(statbuf.st_mode)) {
	if (argc > 2) {
	    Tcl_AppendResult(interp, "target \"", argv[argc - 1],
			     "\" is not a directory", (char *) NULL);
	    Tcl_DStringFree(&targetBuffer) ;
	    return TCL_ERROR;
	}
	r = copyfile(interp, argv[0], targetFileName, force) ;
	Tcl_DStringFree(&targetBuffer) ;
	
	return(r) ;
    }
    
    /*
     * If the target is a directory, then copy each file into it.
     */
    
    Tcl_DStringInit(&buffer) ;

    for (i = 0; i < argc - 1; i++ ) {
	/*
	 * Extract the basename from each source, and
	 * append it to the end of the target path.
	 */
	Tcl_SplitPath(argv[i], &pargc, &pargv) ;
	jargv = (char **)ckalloc(2*sizeof(char *)) ;
	jargv[0] = targetFileName ;
	jargv[1] = pargv[pargc - 1] ;
	
	filename = Tcl_JoinPath(2, jargv, &buffer) ;
	
	ckfree((char *)pargv) ;
	ckfree((char *)jargv) ;
	
	r = copyfile(interp, argv[i], filename, force) ;

	Tcl_DStringFree(&buffer) ;
	if (r == TCL_ERROR) {
	    Tcl_DStringFree(&targetBuffer) ;
	    return TCL_ERROR;
	}

    }
    Tcl_DStringFree(&targetBuffer) ;
	
    return TCL_OK ;
}

/*
 *----------------------------------------------------------------------
 *
 * copyfile
 *
 *	copy a single file or directory
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *      If source is a nonempty directory, the contents of source will      
 *      be replicated inside a directory named by target
 *----------------------------------------------------------------------
 */

static int
copyfile(interp, source, target, force) 
    Tcl_Interp *interp ;
    char *source ;
    char *target ;
    int force ;
{
    struct stat dstStatBuf, srcStatBuf;
    int r ;
    Tcl_DString sourcePath, targetPath;
    char *targetName, *sourceName ;
    

    Tcl_DStringInit(&sourcePath) ;
    sourceName = Tcl_TranslateFileName(interp, source, &sourcePath) ;
    if (sourceName == NULL) {
	Tcl_AppendResult(interp,"illegal file name \"",
			 source, "\"", (char *)NULL);
	Tcl_DStringFree(&sourcePath);
	return TCL_ERROR;
    }

    Tcl_DStringInit(&targetPath) ;
    targetName = Tcl_TranslateFileName(interp, target, &targetPath) ;
    if (targetName == NULL) {
	Tcl_AppendResult(interp,"illegal file name \"",
			 target, "\"", (char *)NULL);
	r = TCL_ERROR ;
	goto error ;
    }

    if (stat(sourceName, &srcStatBuf) < 0 ) {
	goto error;
    }
    if (stat(targetName, &dstStatBuf) == 0) {
	if (S_ISDIR(srcStatBuf.st_mode) && !S_ISDIR(dstStatBuf.st_mode)) {
	    Tcl_AppendResult(interp, "can't overwrite non-directory \"", 
			     source, "\" with directory \"", target, "\"", 
			     (char *) NULL) ;
	    goto error ;
	}
	if (!S_ISDIR(srcStatBuf.st_mode) && S_ISDIR(dstStatBuf.st_mode)) {
	    Tcl_AppendResult(interp, "can't overwrite directory \"", source, 
			     "\" with non-directory \"", target, "\"", 
			     (char *) NULL) ;
	    goto error ;
	}
    }

    if (!S_ISDIR(srcStatBuf.st_mode)) {
	r = TclCopyFile(sourceName, targetName, force) ;
	if (r == TCL_OK) {
	    goto done;
	}
    } else {
	r = TclCopyDirectory(sourceName, targetName, force) ;
	if (r == TCL_OK) {
	    goto done;
	}
    }
    
error:
    Tcl_AppendResult(interp, "error copying \"", source, 
		     "\" to \"", target, "\" : ", (char *)NULL) ;
    Tcl_AppendResult(interp, Tcl_PosixError(interp), (char *)NULL) ;
    r = TCL_ERROR ;
done:
    Tcl_DStringFree(&sourcePath) ;
    Tcl_DStringFree(&targetPath) ;
    return r ;
}

/*
 *----------------------------------------------------------------------
 *
 * TclMakeDirsCmd
 *
 *	This procedure implements the "mkdir" subcommand of the "file"
 *      command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */
int
TclMakeDirsCmd(interp, argc, argv)
        Tcl_Interp *interp ;
	int argc;
	char *argv[];
{
	Tcl_DString buffer ;
	char *dir ;
	int i, j ;
	char **pargv = NULL;
	int pargc ;
	Tcl_DString target ;
	struct stat sb ;
	int intermediate ;

	Tcl_DStringInit(&buffer) ;
	Tcl_DStringInit(&target) ;
	for (i=0; i<argc; i++) {
	    dir = Tcl_TranslateFileName(interp, argv[i], &buffer) ;
	    if (dir == NULL) {
		Tcl_AppendResult(interp, "illegal file name \"", 
				 argv[i], "\"", (char *)NULL) ;
		goto error;
	    }

	    Tcl_SplitPath(dir, &pargc, &pargv) ;
	    if (pargc == 0) {
		Tcl_AppendResult(interp, "illegal file name \"", 
				 argv[i], "\"", (char *)NULL) ;
		goto error;
	    }
	    intermediate = 1 ;
	    for (j=0; j<pargc; j++) {
		Tcl_JoinPath(j+1, pargv, &target) ;
		
		if (stat(Tcl_DStringValue(&target), &sb)) {
		    if (errno != ENOENT) {
			Tcl_AppendResult(interp, "stat failed on \"", 
					 Tcl_DStringValue(&target), "\" : ",
					 Tcl_PosixError(interp), 
					 (char *)NULL) ;
			goto error;
		    }
		    if (j == (argc - 1)) {
			intermediate = 0 ;
		    }
		    if (TclMkdir(Tcl_DStringValue(&target), intermediate) 
			== TCL_ERROR) {
			Tcl_AppendResult(interp, "mkdir failed on \"", 
					 Tcl_DStringValue(&target), "\" : ", 
					 Tcl_PosixError(interp), 
					 (char *)NULL) ;
			goto error;
		    }
		} else {
		    if (!S_ISDIR(sb.st_mode)) {
			Tcl_AppendResult(interp, "file \"", 
					 Tcl_DStringValue(&target),  
					 "\" exists and is not a directory", 
					 (char *)NULL) ;
		    error:
			Tcl_DStringFree(&buffer) ;
			Tcl_DStringFree(&target) ;
			if (pargv != NULL) {
			    ckfree((char *)pargv) ;
			}
			return TCL_ERROR ;
		    }
		}
		Tcl_DStringFree(&target) ;
	    }
	    Tcl_DStringFree(&buffer) ;
	    ckfree((char *)pargv) ;
	}
	return TCL_OK ;
}

/*
 *----------------------------------------------------------------------
 *
 * TclDeleteFilesCommand
 *
 *	This procedure implements the "delete" subcommand of the "file"
 *      command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

    /* ARGSUSED */
int
TclDeleteFilesCmd(interp, argc, argv)
        Tcl_Interp *interp ;
	int argc;
	char *argv[];
{
	Tcl_DString buffer ;
	char *file ;
	int i = 0, r ;
	struct stat sb ;
	int force = 0;
	
	while (i < argc) {
	    if (*argv[i] == '-') {
		if (strcmp(argv[i], "-force") == 0) {
		    force = 1;
		} else if (strcmp(argv[i], "--") == 0) {
		    argc-- ;
		    argv++ ;
		    break ;
		} else {
		    Tcl_AppendResult(interp, "bad option \"", argv[i], 
				     " should be -force or --", (char *)NULL) ;
		    return TCL_ERROR ;
		}
		argc-- ;
		argv++ ;
	    } else {
		break ;
	    }
	}
	
	if (argc < 1) {
	    Tcl_AppendResult(interp, "wrong # args: should be ",
			     "\"file delete ?options? file ?file ...?\"",
			     (char *)NULL) ;
	    return TCL_ERROR ;
	}
	Tcl_DStringInit(&buffer) ;
	for (i=0; i<argc; i++) {
	    file = Tcl_TranslateFileName(interp, argv[i], &buffer) ;
	    if (file == NULL) {
		Tcl_AppendResult(interp, "illegal file name \"", 
				 argv[i], "\"", (char *)NULL) ;
		goto error;
	    }

	    if (lstat(file, &sb)) {
		Tcl_AppendResult(interp, "can't remove \"", file, "\" : ",
				 Tcl_PosixError(interp), 
				 (char *)NULL) ;
		goto error ;
	    }
	    
	    if (!S_ISLNK(sb.st_mode) && access(file, W_OK)) {
		/*
		 * The file is not writable by us, so we only 
		 * remove it if the -force option is specified
		 */
		if (!force) {
		    Tcl_AppendResult(interp, "can't remove read-only file \"", 
				     argv[i], "\" must specify -force", 
				     (char *)NULL) ;
		    return TCL_ERROR ;
		}
				     
	    }

	    if (S_ISDIR(sb.st_mode)) {
		r = TclRmdir(file, force) ;
		if (r == TCL_ERROR) {
		    if (((force == 0) && (errno == EEXIST)) ||
			(errno == EACCES)) {
			Tcl_AppendResult(interp, 
					 "permission denied or directory non-empty: \"", 
					 argv[i], "\"", (char *)NULL) ;
			goto error ;
		    }
		}
	    } else {
		r = TclDeleteFile(file) ;
	    }
	    
	    if (r == TCL_ERROR) {
		Tcl_AppendResult(interp, "can't remove \"", argv[i], 
				 "\" : ", Tcl_PosixError(interp),
				 (char *)NULL) ;
		
	    error:
		Tcl_DStringFree(&buffer) ;
		return TCL_ERROR ;
	    }
	}
	return TCL_OK ;
}

