/* 
 * tclMacFCmd.c --
 *
 * Implements the Macintosh specific portions of the file manipulation
 * subcommands of the "file" command.
 *
 * Copyright (c) 1996 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tclMacFCmd.c 1.5 96/08/29 04:13:01
 */

#include "tclInt.h"
#include "tclMacInt.h"
#include "tclPort.h"
#include <FSpCompat.h>
#include <MoreFilesExtras.h>
#include <Strings.h>
#include <Errors.h>
#include <FileCopy.h>
#include <DirectoryCopy.h>
#include <Script.h>
#include <string.h>


#define DOTREE_NS     0     /* Stat failed (no stat ) */
#define DOTREE_PRED   1     /* pre-order directory  */
#define DOTREE_POSTD  2     /* post-order directory */
#define DOTREE_F      3     /* regular file */
#define DOTREE_DNR    4     /* couldn't read directory */

/*
 * type is one of the DO_TREE* constants defined above
 */
typedef int (Tcl_DirTraversalProc)(char *src, char *dst, struct stat *sb, 
                                   int type) ;


static int TraversalDelete _ANSI_ARGS_((char *src, char *dst, 
					struct stat *sbPtr, int type)) ;
static int TraversalCopy _ANSI_ARGS_((char *src, char *dst, 
					struct stat *sbPtr, int type)) ;
static int TraverseUnixTree _ANSI_ARGS_((Tcl_DString *sourcePath, Tcl_DString *destPath, 
                               Tcl_DirTraversalProc *traverseProc)) ;
static pascal Boolean CopyErrHandler(
	OSErr error,              /* Error that occured */
    short failedOperation,    /* operation that caused the error */
	short srcVRefNum,         /* volume ref number of source */
	long srcDirID,            /* directory id of source */
	StringPtr srcName,        /* name of source */
	short dstVRefNum,         /* volume ref number of dst */
	long dstDirID,            /* directory id of dst */
	StringPtr dstName         /* name of dst directory */
)  ;
          
          
static int staticForce = 0 ;    /* Needed for the DirectoryCopy callback CopyErrHandler */
                 
/*
 *----------------------------------------------------------------------
 *
 * TclMacRenameFile --
 *
 *      Changes the location and/or name of a file. Before calling this procedure, 
 *      make sure that src and dst are native pathnames, and if both src and
 *      dst exist, 
 *      that either both are directories or both are regular files.
 *      
 *      
 * Results:
 *		A standard Tcl result.
 *
 * Side effects:
 *      A file is renamed. If force != 0 then if the file dst will be 
 *      overwritten if it exists. If an error occurs, the macintosh error is
 * 		converted to a Posix error, and errno is set appropriately.
 *
 *----------------------------------------------------------------------
 */

int
TclMacRenameFile( 
    char *src,     /* native pathname of the existing file to rename */
    char *dst,     /* new native pathname for file */
    int  force     /* if != 0, then existing files will be overwritten */
) {
	FSSpec srcFileSpec, dstFileSpec, newSpec, srcDirSpec;
	OSErr err, err2 ;
	long dstDirID ;
	Boolean dstIsDirectory ;
	char *fileName = NULL ;
	char *dirName ;
	Tcl_DString buffer ;
	int pargc ;
	char **pargv ;

	err = FSpLocationFromPath(strlen(src), src, &srcFileSpec) ;
	if (err != noErr) {
		TclMacOSErrorToPosixError(err) ;
		return TCL_ERROR ;
    }
    err = FSpLocationFromPath(strlen(dst), dst, &dstFileSpec) ;
	if (err == noErr)  {
		err = DirIDFromFSSpec(&dstFileSpec, &dstDirID, &dstIsDirectory) ;
	}
    if (err == noErr) {
    	/*
    	 * Existing files are only overwritten if the "-force"
    	 * option is specified. Existing directories are never
    	 * overwritten, but no error is returned.
    	 * FSpMoveRenameCompat will fail if destination file exists, so 
    	 * the dst must be removed beforehand if necessary
    	 */
    	 
    	 if (!dstIsDirectory) {
    	 	if (force) {
    	 		/*
		 		 * Prevent renaming a file to itself.
		 		 */
		 
				 if (srcFileSpec.parID == dstFileSpec.parID) {
		 			if (strcmp((const char *) p2cstr(srcFileSpec.name), 
				 	 	(const char *) p2cstr(dstFileSpec.name)) == 0) {
						errno = EINVAL ;
						return TCL_ERROR ;
				 	 }
				 	 c2pstr((char *)srcFileSpec.name) ;
     				 c2pstr((char *)dstFileSpec.name) ;
				 }
    	 		err = FSpDeleteCompat(&dstFileSpec) ;
    	 		if (err != noErr) {
    	 			errno = TclMacOSErrorToPosixError(err) ;
    	 			return TCL_ERROR ;
    	 		}
    	 	} else {
    	 		errno = EEXIST ;
    	 		return TCL_ERROR ;
    		}
		 } else {
		   /*
		    * The dst directory already exists. We are supposed
		    * to implement rename such that renaming a dir to an
		    * an existing dir will merge the contents. However
		    * this has not been implemented.
		    */
		 	errno = EEXIST ;
		 	return TCL_ERROR ;
		 }
	} else {
		/* 
		 * An error occured. fnfErr means that the target file does not
		 * exist, which is fine, but we need to extract the containing directory
		 * from the target path
		 */
		 
		if (err != fnfErr) {
			errno = TclMacOSErrorToPosixError(err) ;
			return TCL_ERROR ;
		}
	}
	
	/* 
	 * In the Mac world, you move a file between directories and rename it 
	 * within directories. So we need to extract the name relative to the
	 * tail directory, and also the tail directory
	 */
	 
	Tcl_DStringInit(&buffer);
	Tcl_SplitPath(dst, &pargc, &pargv) ;
	if (pargc == 1) {
		fileName = pargv[0] ;
		dirName = ":" ;
	} else {
		fileName = pargv[pargc - 1] ;
		dirName = Tcl_JoinPath(pargc -1, pargv, &buffer) ;
	}
	
	c2pstr(fileName) ;

	err = FSpLocationFromPath(strlen(dirName), dirName, &newSpec) ;
	if (err == noErr) {
		err = DirIDFromFSSpec(&newSpec, &dstDirID, &dstIsDirectory) ;
	}
	if (err != noErr) {
		goto error;
	}
		
	if (srcFileSpec.parID == dstDirID) {
		err = FSpRename(&srcFileSpec, (StringPtr)fileName) ;
	} else {
   /* 	err = FSpMoveRenameCompat(&srcFileSpec, &newSpec,(StringPtr) fileName);*/
		err = FSpCatMove(&srcFileSpec, &newSpec) ;
		if (err == noErr ) {
			err = DirIDFromFSSpec(&newSpec, &dstDirID, &dstIsDirectory) ;
			if (err == noErr) {
				err = FSMakeFSSpecCompat(newSpec.vRefNum, dstDirID, srcFileSpec.name, 
										&newSpec) ;
			}
			if (err == noErr) {
				err = FSpRename(&newSpec, (StringPtr) fileName) ;
			}
			
			if (err != noErr) {
				/*
				 * If we have an error here, that means that the file has been 
				 * successfully moved, but the rename has not happened. So we
				 * want to move the file back to where it was.
				 */
			 
				err2 = FSMakeFSSpecCompat(newSpec.vRefNum, srcFileSpec.parID, NULL, 
											&srcDirSpec) ;
			 	
			 	if (err2 == noErr) {
					FSpCatMove(&newSpec, &srcDirSpec) ;
				}
				errno = TclMacOSErrorToPosixError(err) ;
				return TCL_ERROR ;
			}
			 
		}
    }
	
	if (err != noErr) {
	error:
		Tcl_DStringFree(&buffer) ;
		ckfree((char *)pargv) ;
		TclMacOSErrorToPosixError(err) ;
		return TCL_ERROR ;
	}

	Tcl_DStringFree(&buffer) ;
	ckfree((char *)pargv) ;
	return TCL_OK ;
}

/*
 *----------------------------------------------------------------------
 *
 * TclMacCopyFile --
 *
 *      Copy a single regular file. Before calling this procedure, 
 *      make sure that src and dst are native pathnames, and if both src and
 *      dst exist, 
 *      that either both are directories or both are regular files.
 *      
 *      
 * Results:
 *		A standard Tcl result.
 *
 * Side effects:
 *      A file is copied. If force != 0, then dst will be overwritten if it 
 *      already exists. If an error occurs, the macintosh error is
 * 		converted to a Posix error, and errno is set appropriately.
 *
 *----------------------------------------------------------------------
 */
int 
TclMacCopyFile(
	char *src,     /* native pathname of existing source file */
 	char *dst,     /* native pathname for new copy */
 	int force      /* if 1, then overwrite existing files */
)
{
	FSSpec srcFileSpec ;
	FSSpec dstFileSpec ;
	FSSpec newSpec ;
	FSSpec dstDirSpec ;
	long dstDirID ;
	Boolean dstIsDirectory ;
	char *fileName = NULL ; 
	OSErr err ;
	char *dirName ;
	Tcl_DString buffer ;
	int pargc ;
	char **pargv ;
	
	err = FSpLocationFromPath(strlen(src), src, &srcFileSpec) ;
	if (err != noErr) {
		errno = TclMacOSErrorToPosixError(err) ;
		return TCL_ERROR ;
    }
    
    err = FSpLocationFromPath(strlen(dst), dst, &dstFileSpec) ;
    if (err == noErr) {
        err = DirIDFromFSSpec(&dstFileSpec, &dstDirID, &dstIsDirectory) ;
        
        /*
         * TclMacCopyFile should not be used to copy directories.
         */
         
        if (dstIsDirectory) {
        	errno = EINVAL;
        	return TCL_ERROR;
        }
    }
    
    if ((err != noErr) && (err != fnfErr)) {
    	errno = TclMacOSErrorToPosixError(err) ;
		return TCL_ERROR ;
	}
	
    if (err == noErr) {
    	/*
    	 * Existing files are only overwritten if the "-force"
    	 * option is specified. Existing directories are never
    	 * overwritten, but no error is returned.
    	 * FSpFileCopy will fail if destination file exists, so 
    	 * the dst must be removed beforehand if necessary
    	 */
    	 
   	 	if (force) {
   	 		/*
	 		 * Copying a file to itself is an error.
	 		 */
	 
			 if (srcFileSpec.parID == dstFileSpec.parID) {
	 			if (strcmp((const char *) p2cstr(srcFileSpec.name), 
			 	 	(const char *) p2cstr(dstFileSpec.name)) == 0) {
			 	 	errno = EEXIST ;
	 			 	return TCL_ERROR ;
			 	 }
			 	 c2pstr((char *)srcFileSpec.name) ;
   				 c2pstr((char *)dstFileSpec.name) ;
			}
   	 		err = FSpDeleteCompat(&dstFileSpec) ;
   	 		if (err != noErr) {
   	 			errno = TclMacOSErrorToPosixError(err) ;
   	 			return TCL_ERROR ;
   	 		}
   	 	} else {
   	 		errno = EEXIST ;
   	 		return TCL_ERROR ;
   		}
	 } 
	 
	 Tcl_DStringInit(&buffer);
	 Tcl_SplitPath(dst, &pargc, &pargv) ;
	 if (pargc == 1) {
	 	fileName = pargv[0] ;
	 	dirName = ":" ;
	 } else {
	 	fileName = pargv[pargc - 1] ;
		dirName = Tcl_JoinPath(pargc -1, pargv, &buffer) ;
	 }
	
	c2pstr(fileName) ;
	err = FSpLocationFromPath(strlen(dirName), dirName, &dstDirSpec) ;	
	if (err == noErr) {
    	    err = DirIDFromFSSpec(&dstDirSpec, &dstDirID, &dstIsDirectory);
   	}
	if (err == noErr) {
	    if (dstIsDirectory) {
	    	err = FSMakeFSSpecCompat(dstDirSpec.vRefNum, dstDirID, NULL, &newSpec);
	    } else {
	    	err = dirNFErr;
	    }
	}
	
	if (err != noErr) {
		goto error ;
	}
    
	/*
	 * The fourth argument specifies a buffer to use while copying the file
	 * Since we specify this as NULL, the system will allocate a buffer on
	 * the application heap. The fifth argument is the size of the buffer, 
	 * if we had passed one. The final argument tells the system to check
	 * first to make sure that enough space exists for the new copy of the 
	 * file before attempting the copy.
	 */
	err = FSpFileCopy(&srcFileSpec, &newSpec, (StringPtr)fileName, NULL, 0, true);

	if (err != noErr) {
	error:
		Tcl_DStringFree(&buffer) ;
		ckfree((char *)pargv) ;
		errno = TclMacOSErrorToPosixError(err) ;
		return TCL_ERROR ;
	}

	Tcl_DStringFree(&buffer) ;
	ckfree((char *)pargv) ;
	
	return TCL_OK ;
}

/*
 *----------------------------------------------------------------------
 *
 * TclMacMkdir --
 *
 *      Creates a directory. Fails if the directory already exists.
 *
 * Results:
 *		A standard Tcl result.
 *
 * Side effects:
 *      A directory is created. If an error occurs, the macintosh error is
 * 		converted to a Posix error, and errno is set appropriately.
 *
 *----------------------------------------------------------------------
 */

int
TclMacMkdir(
    char *path,           /* native pathname of directory to be created */
    int ignore           /* ignored on the macintosh */
)
{
	OSErr err ;
	FSSpec dirSpec ;
	long outDirID ;
	
	err = FSpLocationFromPath(strlen(path), path, &dirSpec);
	if (err == noErr) {
		errno = EEXIST ;
		return TCL_ERROR ;
	} else if (err != fnfErr) {
		errno = TclMacOSErrorToPosixError(err) ;
		return TCL_ERROR ;
	}
	
	/*
	 * The directory ID of the new dir is stored in outDirID.
	 * smSystemScript basically means that we didn't establish
	 * the name of the directory through a standard file dialog.
	 * see "Inside Macintosh: Files" for more information.
	 */
	err = FSpDirCreateCompat(&dirSpec, smSystemScript, &outDirID);
	if (err != noErr) {
		errno = TclMacOSErrorToPosixError(err) ;
		return TCL_ERROR ;
	}

	return TCL_OK ;
}

/*
 *----------------------------------------------------------------------
 *
 * TclMacRmdir --
 *
 *      Removes a directory.
 *
 * Results:
 *      Standard Tcl result.
 *
 * Side effects:
 *      If force is true, then the contents of the specified directory
 *      will also be removed recursively. If an error occurs, the macintosh error is
 * 		converted to a Posix error, and errno is set appropriately.
 *----------------------------------------------------------------------
 */
int
TclMacRmdir(path,force) 
    char *path ;   /* path of directory to be removed */
    int force ;    /* remove directories that are nonempty */
{
	OSErr err ;
	FSSpec fileSpec ;
	long dirID ;
	Boolean isDirectory ;
	
	err = FSpLocationFromPath(strlen(path), path, &fileSpec);
	if (err != noErr) {
		errno = TclMacOSErrorToPosixError(err) ;
		return TCL_ERROR ;
	}

	/*
	 * fBsyErr means one of three things: file busy, directory not empty, 
	 * or working directory control block open. We assume that the
	 * directory is not empty, and if that is wrong, then an error from
	 * the next system call should break us out of TclMacRmdir.
	 */
	 
	err = FSpDeleteCompat(&fileSpec) ;
	if ( err == noErr ) {
		return TCL_OK ;
	} else if (err != fBsyErr) {
		errno = TclMacOSErrorToPosixError(err) ;
		return TCL_ERROR ;
	}
	
	/*
	 * We got an fBsyErr, which we are assuming means that the directory is
	 * not empty
	 */
	 
	if (!force) {
		errno = EEXIST ;
		return TCL_ERROR ;
	}
	
	err = DirIDFromFSSpec(&fileSpec, &dirID, &isDirectory) ;
	if (err != noErr) {
		errno = TclMacOSErrorToPosixError(err) ;
		return TCL_ERROR ;
	}
	
	/*
	 * DeleteDirectory removes a directory and all its contents.
	 * DeleteDirectory will unlock and remove any locked files 
	 * that it finds. This really shouldn't happen unless the
	 * "-force" option is specified, but there is no way around 
	 * it without rewriting the MoreFiles DeleteDirectory procedure.
	 * The last argument is the name of the directory. Since dirID refers
	 * to a directory, then we don't need to specify it.
	 */
	 
	err = DeleteDirectory(fileSpec.vRefNum, dirID, NULL) ;
	if (err != noErr) {
		errno = TclMacOSErrorToPosixError(err) ;
		return TCL_ERROR ;
	}
	return TCL_OK ;
}

/*
 *----------------------------------------------------------------------
 *
 * TclMacDeleteFile
 *
 *      Removes a single file. 
 *
 * Results:
 *      Standard Tcl result.
 *
 * Side effects:
 *      The file specified by path is removed. If an error occurs, 
 * 		the macintosh error is converted to a Posix error, 
 *      and errno is set appropriately.
 *----------------------------------------------------------------------
 */

int TclMacDeleteFile(
    char *path     /* native pathname of file to be removed */
)
{
	OSErr err ;
	FSSpec fileSpec ;
	
	err = FSpLocationFromPath(strlen(path), path, &fileSpec);
	if (err != noErr) {
		errno = TclMacOSErrorToPosixError(err) ;
		return TCL_ERROR ;
	}

	err = FSpDeleteCompat(&fileSpec) ;
	if ( err != noErr ) {
		errno = TclMacOSErrorToPosixError(err) ;
		return TCL_ERROR ;
	}
	
  	return TCL_OK ;
}

/*
 *----------------------------------------------------------------------
 *
 * TclMacCopyDirectory --
 *
 *      The file hierarchy rooted at src will be copied to a file hierarchy
 *      rooted at dst. If dst exists, it must be a directory.
 *
 * Results:
 *      Standard Tcl result.
 *
 * Side effects:
 *      A directory and its contents are copied. If force != 0 then
 *      existing files will be overwritten.If an error occurs, 
 * 		the macintosh error is converted to a Posix error, 
 *      and errno is set appropriately.
 *      
 *----------------------------------------------------------------------
 */
int
TclMacCopyDirectory(
    char *src,    /* Pathname of directory to be copied */
    char *dst,    /* Pathname of target directory */
    int force     /* if != 0, then overwrite existing files */
)
{
	OSErr err ;
	FSSpec srcDirSpec, dstDirSpec, targetDirSpec , tmpDirSpec, tmpTargetSpec;
	long targetDirID , tmpDirID;
	Boolean dstIsDirectory ;
	char *fileName ;
	char *dirName ;
	int pargc ;
	char **pargv = NULL;
	int targetExists = 0;
	Tcl_DString buffer ;
		
	err = FSpLocationFromPath(strlen(src), src, &srcDirSpec);
	if (err != noErr) {
		errno = TclMacOSErrorToPosixError(err) ;
		return TCL_ERROR ;
	}

	err = FSpLocationFromPath(strlen(dst), dst, &dstDirSpec);
	if ((err != noErr) && (err != fnfErr)) {
		errno = TclMacOSErrorToPosixError(err) ;
		return TCL_ERROR ;
	}
	/* 
	 * The MoreFiles procedure FSpDirectoryCopy copies a src dir
	 * INTO a dst dir. We want to copy the contents of src into
	 * dst. So we need to find the parent of dst and copy src 
	 * into that instead. Then, since FSpDirectoryCopy doesn't rename
	 * a the directory, we have to call rename on it.
	 */
	Tcl_DStringInit(&buffer);
	if (err == noErr) {
		targetExists = 1 ;
		targetDirID = dstDirSpec.parID ;
	} else {
		Tcl_SplitPath(dst, &pargc, &pargv) ;
		if (pargc == 1) {
			fileName = pargv[0] ;
			dirName = ":" ;
		} else {
			fileName = pargv[pargc - 1] ;
			dirName = Tcl_JoinPath(pargc -1, pargv, &buffer) ;
		}
		err = FSpLocationFromPath(strlen(dirName), dirName, &dstDirSpec) ;
		if (err == noErr)  {
			DirIDFromFSSpec(&dstDirSpec, &targetDirID, &dstIsDirectory) ;
		}
		if (err != noErr) {
			goto error ;
		}
	}
	
	err = FSMakeFSSpecCompat(dstDirSpec.vRefNum, targetDirID, NULL, &targetDirSpec) ;
	if (err != noErr) {
		goto error ;
	}
	
	/*
	 * FSpDirectoryCopy calls CopyErrHandler for every error that it finds.
	 * CopyErrHandler needs to access the value of force, but FSpDirectoryCopy()
	 * provides no way to pass data to that function. So we need to provide a global
	 * variable to allow CopyErrHandler to access it. This is not threadsafe
	 * at all. A good compromise would be to have a hashtable that is indexed
	 * on the interpreter name. But we would need to pass interp into 
	 * TclMacCopyDirectory()
	 */
	staticForce = force ;
	if ((srcDirSpec.parID == targetDirID) && !targetExists) {
		/*
		 * Both the source and the dest are in the same directory.
		 * However, macos and morefiles do not provide any directory copy routine 
		 * that allows both a copy and a rename, so copying a directory
		 * where the source and target are in the same directory is not
		 * possible. There is a lot of mumbo-jumbo that the FSpDirectoryCopy
		 * function does while traversing the directory tree that we do not 
		 * want to duplicate, so reimplimenting FSpDirectoryCopy is not a
		 * good option. So instead, we use a two stage approach where we create 
		 * a temporary directory where the target will reside, copy the src into
		 * that temp directory, then move the new directory back up one level 
		 * renaming it in the process.
		 */
		 
		 /*
		  * First we need to create a directory that has a high probability
		  * of being unique. We take the target dir's name and append the current
		  * directory's id.
		  */
		  
		  Tcl_DString tmpName;
		  char dirString[25] ;
		  
		  Tcl_DStringInit(&tmpName);
		  Tcl_DStringAppend(&tmpName, fileName, -1) ;
		  sprintf(dirString, "%d", targetDirID) ;
		  Tcl_DStringAppend(&tmpName, dirString, -1) ;
		  c2pstr(Tcl_DStringValue(&tmpName));
		  c2pstr(fileName) ;
		  err = FSMakeFSSpecCompat(srcDirSpec.vRefNum, targetDirID, (StringPtr) Tcl_DStringValue(&tmpName),
		  							 &tmpDirSpec);
		  if ((err == fnfErr) || (err == noErr)) {
		  		err = FSpDirCreateCompat(&tmpDirSpec, smSystemScript, &tmpDirID);
		  }
		  
		  if (err == noErr) {
		  		err = FSpDirectoryCopy(&srcDirSpec, &tmpDirSpec, NULL, 0, true, CopyErrHandler) ;
		  } else {
		  		/*
		  		 * Need to separate errors happening before the copy from errors
		  		 * happening after.
		  		 */
		  		
		  		goto error;
		  }
		  		 
		  if (err == noErr) {
		  		err = FSMakeFSSpecCompat(srcDirSpec.vRefNum, tmpDirID, srcDirSpec.name,
		  							 &tmpTargetSpec);
		  }
		  if (err == noErr) {
		  		err = FSpRename(&tmpTargetSpec, (StringPtr)fileName) ;
		  }

		  if (err == noErr) {
		  		err = FSMakeFSSpecCompat(srcDirSpec.vRefNum, tmpDirID, (StringPtr)fileName,
		  							 &tmpTargetSpec);
		  }
		  if (err == noErr) {
		  		err = FSpCatMove(&tmpTargetSpec, &targetDirSpec) ;
		  }
		  if (err == noErr) {
		  		err = FSpDeleteCompat(&tmpDirSpec) ;
		  }
		  
		  if (err != noErr) {
		  		FSpDeleteCompat(&tmpDirSpec) ;
		  		goto error;
		  }
	} else {
		err = FSpDirectoryCopy(&srcDirSpec, &targetDirSpec, NULL, 0, true, CopyErrHandler) ;
		staticForce = 0 ;
		if (err != noErr) {
		error:
			Tcl_DStringFree(&buffer) ;
			if (pargv != NULL) {
				ckfree((char *)pargv) ;
			}
			errno = TclMacOSErrorToPosixError(err) ;
			return TCL_ERROR ;
		}
	}
	Tcl_DStringFree(&buffer);
	if (pargv != NULL) {
		ckfree((char *)pargv) ;
	}
	return TCL_OK ;
}


/*
 *----------------------------------------------------------------------
 *
 * CopyErrHandler --
 *
 *      This procedure is called from the MoreFiles procedure 
 *		FSpDirectoryCopy whenever an error occurs.
 *
 * Results:
 *      false if the condition should not be considered an error, true
 *      otherwise.
 *
 * Side effects:
 *      If FSpDirectoryCopy tries to create a directory that already exists, 
 *      the error is ignored. If FSpDirectoryCopy tries to overwrite an 
 * 		existing file, then correct action is taken based on the value 
 *      of the "-force" option.
 *      
 *----------------------------------------------------------------------
 */
static pascal Boolean 
CopyErrHandler(
	OSErr error,              /* Error that occured */
    short failedOperation,    /* operation that caused the error */
	short srcVRefNum,         /* volume ref number of source */
	long srcDirID,            /* directory id of source */
	StringPtr srcName,        /* name of source */
	short dstVRefNum,         /* volume ref number of dst */
	long dstDirID,            /* directory id of dst */
	StringPtr dstName         /* name of dst directory */
) {
	OSErr err ;
	FSSpec dstSpec, srcSpec ;
	long dummy ;
	Boolean srcIsDirectory, dstIsDirectory ;	
	
	switch (failedOperation) {

	case dirCreateOp:
	 	/*
	 	 * Handle errors where the DirCreate system call failed.
	 	 * If DirCreate() is called on an existing directory, we don't 
	 	 * want to fail. Unfortunately, if we return false, then DirectoryCopy
	 	 * isn't smart enough to open the directory that it just tried to create.
	 	 */
	 	 return true ;
		 break ;
		 
	case copyDirFMAttributesOp:
	case getNextItemOp:
	
	 	/*
	 	 * If an error occurs in the CopyFileMgrAttributes() or PBGetCatInfo() 
	 	 * system calls, then we
	 	 * bail out with an error 
	 	 */
	 	 	  
		return true ;
		break ;
		
	case fileCopyOp:
		/*
		 * This means that the system call FileCopy() failed.
		 */
		 
		 if (error == dupFNErr) {
		 	/* 
		 	 * If the "-force" option has been supplied, and both the src 
		 	 * and dst are regular files, then we want to remove the 
		 	 * dst file, copy it, and then continue the copy.
		 	 */
		 	 
		 	 	err = FSMakeFSSpecCompat(srcVRefNum, srcDirID, srcName, &srcSpec) ;
		 	 	if (err != noErr) {
		 	 		return true ;
		 	 	}
		 	 	err = DirIDFromFSSpec(&srcSpec, &dummy, &srcIsDirectory) ;
		 	 	if (err != noErr) {
		 	 		return true ;
		 	 	}
		 	 	err = FSMakeFSSpecCompat(dstVRefNum, dstDirID, dstName, &dstSpec) ;
		 	 	if (err != noErr) {
		 	 		return true ;
		 	 	}
		 	 	err = DirIDFromFSSpec(&dstSpec, &dummy, &dstIsDirectory) ;
				if (err != noErr) {
		 	 		return true ;
		 	 	}
				if (srcIsDirectory) {
					if (!dstIsDirectory) {
						return true ;
					} else {
						return false ;
					}
				} else {
					if (!dstIsDirectory) {
						return true ;
					} else {
						if (staticForce) {
							err = FSpDeleteCompat(&dstSpec) ;
		 	 				if (err != noErr) {
		 	 					return true ;
		 	 				}
		 	 				err = FSpFileCopy(&srcSpec, &dstSpec, srcName, NULL, 0, true);
							if (err != noErr) {
		 	 					return true ;
		 	 				}
		 	 				return false ;
						} 
		 	 		}
		 	 	}
		 	 	
		 
		

		} else {
			return true ;
		}
		break ;
	}

	return true ;	
}	
