/*
 * tclUnixFCmd.c
 *
 *      This file implements the unix specific portion of file manipulation 
 *      subcommands of the "file" command. 
 *
 * Copyright (c) 1996 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tclUnixFCmd.c 1.8 96/08/29 04:08:27
 */

/*
 * Portions of this code were derived from NetBSD source code which has
 * the following copyright notice:
 *
 * Copyright (c) 1988, 1993, 1994
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "tclInt.h"
#include "tclPort.h"

/*
 * The following constants specify the type of callback when
 * TraverseDirectoryTree() calls the traverseProc()
 */

#define DOTREE_NS     0     /* Stat failed (no stat ) */
#define DOTREE_PRED   1     /* pre-order directory  */
#define DOTREE_POSTD  2     /* post-order directory */
#define DOTREE_F      3     /* regular file */
#define DOTREE_DNR    4     /* couldn't read directory */

/*
 * This is a prototype of the callback function called from 
 * TraverseDirectoryTree. 
 */

typedef int (Tcl_DirTraversalProc) _ANSI_ARGS_((char *src, char *dst, 
	         struct stat *sb, int type, int force)) ;

/*
 * Declarations for local procedures defined in this file:
 */

static int            TraversalDelete _ANSI_ARGS_((char *src, char *dst, 
		          struct stat *sbPtr, int type, int force)) ;
static int            TraversalCopy _ANSI_ARGS_((char *src, char *dst, 
		          struct stat *sbPtr, int type, int force)) ;
static int            TraverseUnixTree _ANSI_ARGS_((Tcl_DString *sourcePath, 
		          Tcl_DString *destPath, 
			  Tcl_DirTraversalProc *traverseProc, int force)) ;
static int            CopyFileAtts _ANSI_ARGS_((char *src, char *dst, 
			  struct stat *statBufPtr));

/*
 *----------------------------------------------------------------------
 *
 * TclUnixRenameFile --
 *
 *      Changes the location and/or name of a file. If both the src and
 *      the dst exist, then they should be both of the same file type
 *      upon entering this function.
 *      
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *      A file is renamed. 
 *
 *----------------------------------------------------------------------
 */

int
TclUnixRenameFile(src, dst, force) 
    char *src; 
    char *dst;
    int force ;
{
	struct stat ssb, tsb;
	int r ;

	r = stat(src, &ssb) ;
	if (!r) {
	    r = stat(dst, &tsb) ;
	    if (!r) {
		if (S_ISDIR(tsb.st_mode)) {
		    /*
		     * Renaming a directory to an existing directory
		     * is supposed to merge the contents. This is not 
		     * currently supported.
		     */
		    errno = EEXIST ;
		    return TCL_ERROR ;
		}
		if (force) {
		    /*
		     * Dst file already exists
		     */
		    if ((ssb.st_dev == tsb.st_dev) && (ssb.st_ino == tsb.st_ino)) {
			/*
			 * Files are identical, return an error 
			 */
			errno = EINVAL ;
			return TCL_ERROR ;
		    }
		} else {
		    errno = EEXIST ;
		    return TCL_ERROR;
		}
	    } 
	} 

	if (rename(src, dst)) {
	    return TCL_ERROR ;
	}
	return TCL_OK ;
}


/*
 *----------------------------------------------------------------------
 *
 * TclUnixCopyFile --
 *
 *      Copy a single file.
 *      
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *      A file is copied.
 *
 *----------------------------------------------------------------------
 */

int 
TclUnixCopyFile(src, dst, force) 
 	char *src, *dst ;
        int force; 
{
 	u_int blen; 
	char *bp;
	int nread, from_fd, to_fd;
	struct stat srcStatBuf, dstStatBuf ;

	if (stat(src, &srcStatBuf) == 0) {
	    if (stat(dst, &dstStatBuf) == 0) {
		if (force) {
		    if ((srcStatBuf.st_dev == dstStatBuf.st_dev) && 
			(srcStatBuf.st_ino == dstStatBuf.st_ino)) {
			/*
			 * Source and target are the same, which is
			 * not an error 
			 */
			return TCL_OK ;
		    }
		} else {
		    errno = EEXIST ;
		    return TCL_ERROR ;
		}
	    }
	} else {
	    return TCL_ERROR ;
	}

 	if ((from_fd = open(src, O_RDONLY, 0)) < 0) { 
	    return TCL_ERROR ;
	}

	if (stat(src, &srcStatBuf) < 0) {
	    return TCL_ERROR ;
	}
	to_fd = open(dst, O_CREAT | O_TRUNC | O_WRONLY, srcStatBuf.st_mode) ;
	if (to_fd < 0) {
	    (void)close(from_fd); 
	    return TCL_OK ;
	}

	blen = srcStatBuf.st_blksize ;
 	if (!(bp = ckalloc(blen))) { 
	    return TCL_ERROR;
 	} 


	while ((nread = read(from_fd, bp, blen)) > 0) {
	    if (write(to_fd, bp, (size_t) nread) != nread) { 
		goto err; 
	    }
	}
	ckfree(bp) ;
	
	if (nread < 0) { 
err:        
	    unlink(dst);

            (void)close(from_fd); 
 	    (void)close(to_fd); 
	    ckfree(bp) ;
	    return TCL_ERROR ;
 	}
	(void)close(from_fd); 

 	if (close(to_fd)) { 
	    return TCL_ERROR ;
 	} 

	if (CopyFileAtts(src, dst, &srcStatBuf) == TCL_ERROR) {
	    /*
	     * the copy succeeded, but setting the permissions
	     * failed, so be in a consistent state, we remove the file
	     * that was created by the copy
	     */
	    TclUnixDeleteFile(dst) ;
	    return TCL_ERROR ;
	}


	return TCL_OK ;
}

/*
 *----------------------------------------------------------------------
 *
 * TclUnixMkdir --
 *
 *      Creates a directory in the filesytem using the mkdir() call.
 *      Permissions of the directory are set based on the current
 *      value of the user's umask.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *      A directory is created.
 *
 *----------------------------------------------------------------------
 */

int
TclUnixMkdir(path, intermediate) 
    char *path ;           /* name of directory to create */
    int intermediate ;     /* 1 if this is an intermediate dir created by */
                           /* TclMkdirCmd */
{
    mode_t mode ;
    mode_t origUmask ;

    /*
     * The default file mode is a=rwx (0777) with selected permissions
     * removed in accordance with the file mode creation mask.  For
     * intermediate path name components, the mode is the default modified
     * by u+wx so that the subdirectories can always be created.
     */


    /*
     * The umask() function both sets the umask and returns the 
     * current value of the umask.
     */
    origUmask = umask(0) ;
    mode = 0777 & ~origUmask;
    umask(origUmask) ;

    if (intermediate) {
	mode = mode | S_IWUSR | S_IXUSR;
    }
    
    if (mkdir(path, mode)) {
	return TCL_ERROR ;
    }
    return TCL_OK ;
}

/*
 *----------------------------------------------------------------------
 *
 * TclUnixRmdir --
 *
 *      Removes directories.
 *
 * Results:
 *      Standard Tcl result.
 *
 * Side effects:
 *      If force is true, then the contents of the specified directory
 *      will be removed recursively.
 *----------------------------------------------------------------------
 */
int
TclUnixRmdir(path,force) 
    char *path ;   /* path of directory to be removed */
    int force ;    /* remove directories that are nonempty */
{
    int r ;
    Tcl_DString buffer ;

    if (rmdir(path) != 0) {
	if (((errno != EEXIST) && (errno != ENOTEMPTY)) || !force) {
	    return TCL_ERROR ;
	}
    } else {
	return TCL_OK ;
    }
    
    /*
     * The directory is nonempty, but the -force flag has been 
     * specified, so we recursively remove all the files in the 
     * directory
     */
    Tcl_DStringInit(&buffer) ;
    Tcl_DStringAppend(&buffer, path, -1) ;
    r = TraverseUnixTree(&buffer, NULL, TraversalDelete, force); 
    Tcl_DStringFree(&buffer) ;
    return r ;
}

/*
 *----------------------------------------------------------------------
 *
 * TclUnixDeleteFile --
 *
 *      Unlinks a file (not a directory!)
 *
 * Results:
 *      Standard Tcl result.
 *
 * Side effects:
 *      file is unlinked.
 *
 *----------------------------------------------------------------------
 */
int TclUnixDeleteFile(path) 
    char *path ;    /* pathname of file to be removed */
{
    if (unlink(path) != 0) {
	return TCL_ERROR ;
    }
    return TCL_OK ;
}
	
/*
 *----------------------------------------------------------------------
 *
 * TraverseUnixTree --
 *
 *      Traverse a directory tree, calling the function traverseProc()
 *      for each file and directory encountered. If destPath is non-null,
 *      the filename relative to sourcePath will be appended to destPath
 *      and passed as the second argument to traverseProc() .
 *
 * Results:
 *      Standard Tcl result.
 *
 * Side effects:
 *      None caused by TraverseUnixTree, however the user specified traverseProc() 
 *      may change state.
 *
 *----------------------------------------------------------------------
 */

static int 
TraverseUnixTree(sourcePath, destPath, traverseProc, force)
    Tcl_DString *sourcePath ;
    Tcl_DString *destPath ;
    Tcl_DirTraversalProc *traverseProc ;
    int force ;
{
    struct stat statbuf ;
    struct dirent *dirp ;
    DIR *dp ;

    int ret = TCL_OK  ;
    char *ptr ;
    char *fullpath, *targetPath ;
    int sourceLen;
    int destLen = 0;			/* Initialization needed only to
					 * prevent gcc compiler warning. */

    fullpath = Tcl_DStringValue(sourcePath) ;
    if (destPath != NULL) {
	targetPath = Tcl_DStringValue(destPath) ;
    } else {
	targetPath = NULL ;
    }

    if (lstat(fullpath, &statbuf) < 0) {
	return(traverseProc(fullpath, targetPath, &statbuf, DOTREE_NS,force)) ;
    }
    if (S_ISDIR(statbuf.st_mode) == 0) {
	/*
	 * Process the regular file
	 */
	return(traverseProc(fullpath, targetPath, &statbuf, DOTREE_F, force)) ;
    }
    /*
     * It's a directory, first call traverseProc() on directory
     * for preorder traversals, then 
     * process the files in that directory.
     */
    ret = traverseProc(fullpath, targetPath, &statbuf, DOTREE_PRED, force) ;
    if ( ret != TCL_OK) {
	return(ret) ;
    }

    if ((dp = opendir(fullpath)) == NULL) {
	/* 
	 * Can't read directory
	 */
	return(traverseProc(fullpath, targetPath, &statbuf, DOTREE_DNR,force)) ;
    }
    
    ptr = fullpath + Tcl_DStringLength(sourcePath) ;
    Tcl_DStringAppend(sourcePath, "/", 1) ;
    fullpath = Tcl_DStringValue(sourcePath) ;

    if (destPath != NULL) {
	Tcl_DStringAppend(destPath, "/", 1) ;
	targetPath = Tcl_DStringValue(destPath) ;
    }
	    
    sourceLen = Tcl_DStringLength(sourcePath) ;	
    if (destPath != NULL) {
	destLen = Tcl_DStringLength(destPath) ;
    }
				  
    while ( (dirp = readdir(dp)) != NULL) {
	if (strcmp(dirp->d_name, ".") == 0 ||
	    strcmp(dirp->d_name, "..") == 0) {
	    continue ;
	}
	/* 
	 * append name after slash 
	 */
	Tcl_DStringAppend(sourcePath, dirp->d_name, -1) ;
	if (destPath != NULL) {
	    Tcl_DStringAppend(destPath, dirp->d_name, -1) ;
	}
	ret = TraverseUnixTree(sourcePath, destPath, traverseProc, force);
	if ( ret != TCL_OK) {
	    break ;
	}
	Tcl_DStringSetLength(sourcePath, sourceLen) ;
	if (destPath != NULL) {
	    Tcl_DStringSetLength(destPath, destLen) ;
	}
    }
    
    /*
     * Strip off the trailing slash
     */
    if (ret == TCL_OK) {
	Tcl_DStringSetLength(sourcePath, sourceLen - 1) ;
	if (destPath != NULL) {
	    Tcl_DStringSetLength(destPath, destLen - 1) ;
	}
	
	/*
	 * Call traverseProc() on a directory after visiting all the
	 * files in that directory.
	 */
	ret = traverseProc(fullpath, targetPath, &statbuf, DOTREE_POSTD,
			   force) ;
    }
    if (closedir(dp) != 0) {
	ret = TCL_ERROR ;
    }
    return ret ;
}


/*
 *----------------------------------------------------------------------
 *
 * TraversalDelete --
 *
 *      Called by procedure TraverseUnixTree for every file and directory that
 *      it encounters in a directory hierarchy. This procedure unlinks
 *      files, and removes directories after all the containing files 
 *      have been processed.
 *
 * Results:
 *      Standard Tcl result.
 *
 * Side effects:
 *      Files or directory specified by src will be deleted.
 *
 *----------------------------------------------------------------------
 */

static int
TraversalDelete(src, ignore, sbPtr, type, force) 
    char *src ;           /* source pathname */
    char *ignore ;        /* destination pathname is ignored for delete op */
    struct stat *sbPtr ;   /* stat info for file specified by src */
    int type ;            /* type of file or operation */
    int force ;
{
    switch (type) {
    case DOTREE_F:
	/*
	 * Not a directory.
	 */
	return (TclUnixDeleteFile(src)) ;
	break  ;
    case DOTREE_PRED:
	return TCL_OK ;
	break ;
    case DOTREE_POSTD:
	return (TclUnixRmdir(src, 0)) ;
	break ;
    case DOTREE_DNR:
    case DOTREE_NS:
	return TCL_ERROR ;
	break ;
    }
    return TCL_ERROR ;
}


/*
 *----------------------------------------------------------------------
 *
 * TclUnixCopyDirectory --
 *
 *      Recursively copies a directory
 *
 * Results:
 *      Standard Tcl result.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */
int
TclUnixCopyDirectory(src, dst, force)
    char *src ;   /* Pathname of directory to be copied */
    char *dst ;   /* Pathname of target directory */
    int force ;
{
    int r ;
    Tcl_DString srcBuffer ;
    Tcl_DString dstBuffer ;

    Tcl_DStringInit(&srcBuffer) ;
    Tcl_DStringInit(&dstBuffer) ;
    Tcl_DStringAppend(&srcBuffer, src, -1) ;
    Tcl_DStringAppend(&dstBuffer, dst, -1) ;
    r = TraverseUnixTree(&srcBuffer, &dstBuffer, TraversalCopy, force) ;
    Tcl_DStringFree(&srcBuffer) ;
    Tcl_DStringFree(&dstBuffer) ;
    return r ;
}


/*
 *----------------------------------------------------------------------
 *
 * TraversalCopy
 *
 *      Called from TraverseUnixTree in order to execute a recursive copy of a 
 *      directory.
 *
 * Results:
 *      Standard Tcl result.
 *
 * Side effects:
 *      Files and directories are copyied.
 *      
 *----------------------------------------------------------------------
 */
static int 
TraversalCopy(src, dst, srcStatBufPtr, type, force) 
    char *src ;
    char *dst ;
    struct stat *srcStatBufPtr;
    int type ;
    int force ;
{
    struct stat dstStatBuf ;

    switch (type) {
    case DOTREE_F:
	/*
	 * Not a directory.
	 */
	return (TclUnixCopyFile(src, dst, force)) ;
	break  ;
    case DOTREE_PRED:
	if (stat(dst, &dstStatBuf) < 0) {
	    if (errno != ENOENT) {
		return TCL_ERROR  ;
	    }
	    if (TclUnixMkdir(dst, 0) == TCL_ERROR) {
		return TCL_ERROR ;
	    }
	}
	if (CopyFileAtts(src, dst, srcStatBufPtr) == TCL_ERROR) {
	    return TCL_ERROR ;
	}

	return TCL_OK ;
	break ;
    case DOTREE_POSTD:
	return TCL_OK ;
	break ;
    case DOTREE_DNR:
    case DOTREE_NS:
	return TCL_ERROR ;
	break ;
    }
    return TCL_ERROR ;
}


/*
 *----------------------------------------------------------------------
 *
 * CopyFileAtts
 *
 *      Copy the file attributes such as owner, group, permissions, and
 *      modification date from one file to another.
 *
 * Results:
 *      Standard Tcl result.
 *
 * Side effects:
 *      user id, group id, permission bits, last modification time, and 
 *      last access time are updated in the new file to reflect the old
 *      file.
 *      
 *----------------------------------------------------------------------
 */
static int
CopyFileAtts(src, dst, statBufPtr) 
    char *src ;                 /* Path name of source file */
    char *dst ;                 /* Path name of target file */
    struct stat *statBufPtr ;   /* ptr to stat info for source file */
{
 	struct timeval tval[2]; 

     	if (chown(dst, statBufPtr->st_uid, statBufPtr->st_gid)) {
	    if (errno == EPERM) {
		/*
		 * Some Unix systems only allow root to use the
		 * chown() function. According to Posix, you can
		 * check this ahead of time by calling pathconf() with
		 * the _PC_CHOWN_RESTRICTED flag, but I am not
		 * just just how portable pathconf() is.
		 */

		return TCL_OK ;
	    }

	    return TCL_ERROR ;
	}

 	if (chmod(dst, statBufPtr->st_mode)) {
	    return TCL_ERROR ;
	}

	tval[0].tv_sec = statBufPtr->st_atime; 
 	tval[1].tv_sec = statBufPtr->st_mtime; 
 	tval[0].tv_usec = tval[1].tv_usec = 0; 
 	if (utimes(dst, tval)) {
	    return TCL_ERROR ;
	}
	return TCL_OK ;
}
