/*
 * tclWinFCmd.c
 *
 *      This file implements the Windows specific portion of file manipulation 
 *      subcommands of the "file" command. 
 *
 * Copyright (c) 1996 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tclWinFCmd.c 1.6 96/08/29 04:09:11
 */

#include <sys/stat.h>
#include <direct.h>
#include "tclInt.h"
#include "tclPort.h"

#define DOTREE_NS     0     /* Stat failed (no stat ) */
#define DOTREE_PRED   1     /* pre-order directory  */
#define DOTREE_POSTD  2     /* post-order directory */
#define DOTREE_F      3     /* regular file */
#define DOTREE_DNR    4     /* couldn't read directory */

/*
 * Type is one of the DO_TREE* constants defined above.
 */

typedef int (Tcl_DirTraversalProc)(char *src, char *dst, int type,
		int force);


static int		TraverseWinTree _ANSI_ARGS_((
			    Tcl_DString *sourcePath, Tcl_DString *destPath, 
			    Tcl_DirTraversalProc *traverseProc, int force));
static int		TraversalCopy _ANSI_ARGS_((char *src, char *dst,
			    int type, int force));
static int		TraversalDelete _ANSI_ARGS_((char *src, char *dst,
			    int type, int force));

/*
 *----------------------------------------------------------------------
 *
 * TclWinRenameFile --
 *
 *      Changes the location and/or name of a file on a volume.
 *      If both src and dst exists, make sure that they are either 
 *      both non-directories or both directories.
 *      
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *      The file source is renamed to target. If an error occurs, the
 *      windows error is converted to a Posix error and errno is set
 *      accordingly.
 *
 *----------------------------------------------------------------------
 */

int
TclWinRenameFile(src, dst, force) 
    char *src; 
    char *dst;
    int force;
{
    struct stat dstStatBuf;

    /*
     * Is there some way to tell if two files are the same?
     */
    if (stat(dst, &dstStatBuf) == 0) {
	if (S_ISDIR(dstStatBuf.st_mode)) {
	    /*
	     * renaming a directory to an existing directory 
	     * is supposed to merge them. However, this is not
	     * currently implemented.
	     */
	    errno = EEXIST ;
	    return TCL_ERROR ;
	}
	if (!force) {
	    errno = EEXIST;
	    return TCL_ERROR;
	} else {
	    /*
	     * The Windows call MoveFile() will not overwrite 
	     * a file, so we need to remove the file first
	     */

	    if (TclWinDeleteFile(dst) == TCL_ERROR) {
		return TCL_ERROR ;
	    }
	}
    }

    /*
     * MoveFile is a Windows SDK procedure. MoveFile() will fail 
     * if you attempt to move files across volumes. We need to make
     * sure that EXDEV is returned in this case.
     */

    if (!MoveFile(src, dst)) {
	TclWinConvertError(GetLastError());
	return TCL_ERROR;
    }
	

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TclWinCopyFile --
 *
 *      Copies a single file on Windows.
 *      
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      A file and it's attributes are copied. If an error occurs, the
 *      windows error is converted to a Posix error and errno is set
 *      accordingly.
 *
 *----------------------------------------------------------------------
 */

int 
TclWinCopyFile(src, dst, force) 
    char *src;
    char *dst;
    int force;
{
    BOOL failIfExists;

    if (force) {
	failIfExists = FALSE;
    } else {
	failIfExists = TRUE ;
    }

    /*
     * MoveFile is a Windows SDK procedure. 
     * CopyFile() will also copy the file attributes
     * from src to dst.
     */

    if (!CopyFile(src, dst, failIfExists)) {
	TclWinConvertError(GetLastError());
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TclWinCopyDirectory --
 *
 *      Recursively copies a directory
 *
 * Results:
 *      Standard Tcl result.
 *
 * Side effects:
 *      A directory and its contents are copied. If an error occurs, the
 *      windows error is converted to a Posix error and errno is set
 *      accordingly.
 *
 *----------------------------------------------------------------------
 */

int
TclWinCopyDirectory(src, dst, force)
    char *src;   /* Pathname of directory to be copied */
    char *dst;   /* Pathname of target directory */
    int force;   /* if != 0, then overwrite existing files */
{
    int r;

    Tcl_DString srcBuffer;
    Tcl_DString dstBuffer;

    Tcl_DStringInit(&srcBuffer);
    Tcl_DStringInit(&dstBuffer);
    Tcl_DStringAppend(&srcBuffer, src, -1);
    Tcl_DStringAppend(&dstBuffer, dst, -1);
    r = TraverseWinTree(&srcBuffer, &dstBuffer, TraversalCopy, force);
    Tcl_DStringFree(&srcBuffer);
    Tcl_DStringFree(&dstBuffer);
    return r;
}

/*
 *----------------------------------------------------------------------
 *
 * TraversalCopy
 *
 *      Called from TraverseUnixTree in order to execute a recursive
 *      copy of a directory.
 *
 * Results:
 *      Standard Tcl result.
 *
 * Side effects:
 *      Files and directories are copyied.
 *      
 *----------------------------------------------------------------------
 */

static int 
TraversalCopy(src, dst, type, force) 
    char *src;
    char *dst;
    int type;
    int force;
{
    struct stat dstStatBuf;

    switch (type) {
	case DOTREE_F:
	    /*
	     * Not a directory.
	     */
	    return (TclWinCopyFile(src, dst, force));

	case DOTREE_PRED:
	    if (stat(dst, &dstStatBuf) < 0) {
		if (errno != ENOENT) {
		    return TCL_ERROR;
		}
		if (TclWinMkdir(dst, 0) == TCL_ERROR) {
		    return TCL_ERROR;
		}
	    }
	    return TCL_OK;

	case DOTREE_POSTD:
	    return TCL_OK;

	case DOTREE_DNR:
	case DOTREE_NS:
	    return TCL_ERROR;
    }
    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * TclWinMkdir --
 *
 *      Creates a directory in the filesytem using the mkdir() call.
 *      Permissions of the directory are set based on the current
 *      value of the user's umask.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *      A directory is created.If an error occurs, the
 *      windows error is converted to a Posix error and errno is set
 *      accordingly.
 *
 *----------------------------------------------------------------------
 */

int
TclWinMkdir(path, intermediate) 
    char *path;           /* name of directory to create */
    int intermediate;     /* 1 if this is an intermediate dir created by */
			  /* TclMkdirCmd */
{
    /*
     * CreateDirectory() is a Win32 SDK procedure. The second argument 
     * specifies a pointer to a security attributes structure.
     * NULL means to assign the default security attributes.
     * The second argument is used only on WinNT.
     */
    if (!CreateDirectory(path, NULL)) {
	TclWinConvertError(GetLastError());
	return TCL_ERROR;
    }   
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TclWinRmdir --
 *
 *      Removes directories.
 *
 * Results:
 *      Standard Tcl result.
 *
 * Side effects:
 *      If force is true, then the contents of the specified directory
 *      will be removed recursively.If an error occurs, the
 *      windows error is converted to a Posix error and errno is set
 *      accordingly.
 *
 *----------------------------------------------------------------------
 */

int
TclWinRmdir(path, force) 
    char *path;   /* path of directory to be removed */
    int force;    /* remove directories that are nonempty */
{
    Tcl_DString buffer;

    /*
     * Both RemoveDirectory and rmdir on windows return "Permission 
     * Denied" when trying to remove a non-empty directory.
     * There is no way to differentiate this from a normal 
     * permission denied, so any time that "-force" option is 
     * specified, we have to open the directory to see if it is
     * really free or not 
     */
    
    if (!force) {
	if (!RemoveDirectory(path)) {
	    TclWinConvertError(GetLastError());
	    return TCL_ERROR;
	}
    } else {
	Tcl_DStringInit(&buffer);
        Tcl_DStringAppend(&buffer, path, -1);
	TraverseWinTree(&buffer, NULL, TraversalDelete, force);
        Tcl_DStringFree(&buffer);
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TclWinDeleteFile --
 *
 *      Unlinks a file (not a directory!)
 *
 * Results:
 *      Standard Tcl result.
 *
 * Side effects:
 *      file is unlinked.If an error occurs, the
 *      windows error is converted to a Posix error and errno is set
 *      accordingly.
 *
 *----------------------------------------------------------------------
 */

int
TclWinDeleteFile(path) 
    char *path;    /* pathname of file to be removed */
{
    if (unlink(path) != 0) {
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TraverseWinTree --
 *
 *      Traverse a directory tree in Windows, calling the function
 *      traverseProc() for each file and directory encountered. If
 *      destPath is non-null, the filename relative to sourcePath will
 *      be appended to destPath and passed as the second argument to
 *      traverseProc() .
 *
 * Results:
 *      Standard Tcl result.
 *
 * Side effects:
 *      None caused by TraverseWinTree, however the user specified
 *      traverseProc() may change state.  If an error occurs, the
 *      windows error is converted to a Posix error and errno is set
 *      accordingly.
 *
 *----------------------------------------------------------------------
 */

static int 
TraverseWinTree(sourcePath, destPath, traverseProc, force)
    Tcl_DString *sourcePath;
    Tcl_DString *destPath;
    Tcl_DirTraversalProc *traverseProc;
    int force;
{
    struct stat statbuf;
    int ret;
    char *fullpath, *targetPath;
    int sourceLen, destLen;
    HANDLE handle;
    WIN32_FIND_DATA data;
    BOOL found;

    fullpath = Tcl_DStringValue(sourcePath);
    if (destPath != NULL) {
	targetPath = Tcl_DStringValue(destPath);
    } else {
	targetPath = NULL;
    }

    if (lstat(fullpath, &statbuf) < 0) {
	return(traverseProc(fullpath, targetPath, DOTREE_NS, force));
    }
    if (S_ISDIR(statbuf.st_mode) == 0) {

	/*
	 * Process the regular file
	 */

	return(traverseProc(fullpath, targetPath, DOTREE_F, force));
    }

    /*
     * It's a directory, first call traverseProc() on directory
     * for preorder traversals, then 
     * process the files in that directory.
     */

    ret = traverseProc(fullpath, targetPath, DOTREE_PRED, force);
    if ( ret != TCL_OK) {
	return(ret);
    }

    sourceLen = Tcl_DStringLength(sourcePath);
    Tcl_DStringAppend(sourcePath, "\\*.*", 4); 
    handle = FindFirstFile(fullpath, &data);
    Tcl_DStringSetLength(sourcePath, sourceLen);

    if (handle == INVALID_HANDLE_VALUE) {
	TclWinConvertError(GetLastError());
	return(traverseProc(fullpath, targetPath, DOTREE_DNR, force));
    }

    Tcl_DStringAppend(sourcePath, "\\", 1);
    fullpath = Tcl_DStringValue(sourcePath);

    if (destPath != NULL) {
	Tcl_DStringAppend(destPath, "\\", 1);
	targetPath = Tcl_DStringValue(destPath);
    }
	    
    sourceLen = Tcl_DStringLength(sourcePath);	
    if (destPath != NULL) {
	destLen = Tcl_DStringLength(destPath);
    }
	
    for (found = 1; found; found = FindNextFile(handle, &data)) {
	if (strcmp(data.cFileName, ".") == 0 ||
		strcmp(data.cFileName, "..") == 0) {
	    continue;
	}

	/* 
	 * Append name after slash 
	 */

	Tcl_DStringAppend(sourcePath, data.cFileName, -1);
	if (destPath != NULL) {
	    Tcl_DStringAppend(destPath, data.cFileName, -1);
	}
	ret = TraverseWinTree(sourcePath, destPath, traverseProc, force);
	if ( ret != TCL_OK) {
	    break;
	}
	Tcl_DStringSetLength(sourcePath, sourceLen);
	if (destPath != NULL) {
	    Tcl_DStringSetLength(destPath, destLen);
	}
    }
    
    
    /*
     * Strip off the trailing slash
     */

    Tcl_DStringSetLength(sourcePath, sourceLen - 1);
    if (destPath != NULL) {
	Tcl_DStringSetLength(destPath, destLen - 1);
    }

    /*
     * Call traverseProc() for post-order traversals
     */

    ret = traverseProc(fullpath, targetPath, DOTREE_POSTD, force);

    FindClose(handle);

    return ret;
}

/*
 *----------------------------------------------------------------------
 *
 * TraversalDelete --
 *
 *      Called by procedure TraverseWinTree for every file and
 *      directory that it encounters in a directory hierarchy. This
 *      procedure unlinks files, and removes directories after all the
 *      containing files have been processed.
 *
 * Results:
 *      Standard Tcl result.
 *
 * Side effects:
 *      Files or directory specified by src will be deleted. If an
 *      error occurs, the windows error is converted to a Posix error
 *      and errno is set accordingly.
 *
 *----------------------------------------------------------------------
 */

static int
TraversalDelete(src, ignore, type, force) 
    char *src;           /* source pathname */
    char *ignore;        /* destination pathname is ignored for delete op */
    int type;            /* type of file or operation */
    int force;
{
    switch (type) {
	case DOTREE_F:
	    /*
	     * Not a directory.
	     */
	    return (TclWinDeleteFile(src));

	case DOTREE_PRED:
	    return TCL_OK;

	case DOTREE_POSTD:
	    if (TclWinRmdir(src, 0) == TCL_ERROR) {
		return TCL_ERROR;
	    } else {
		return TCL_OK;
	    }

	case DOTREE_DNR:
	case DOTREE_NS:
	    return TCL_ERROR;
    }
    return TCL_ERROR;
}
