/* 
 * tclWinPipe.c --
 *
 *	This file implements the Windows-specific pipeline exec functions.
 *
 * Copyright (c) 1996 by Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tclWinPipe.c 1.28 96/08/27 20:14:00
 */

#include "tclInt.h"
#include "tclPort.h"

#include <dos.h>
#include <fcntl.h>
#include <io.h>
#include <windows.h>
#include <sys/stat.h>

#include "tclWinInt.h"

static BOOL		HasConsole _ANSI_ARGS_((void));
static int		ApplicationType _ANSI_ARGS_((const char *fileName));
static BOOL 		IsWinApp _ANSI_ARGS_((char *filename,
			    Tcl_Interp *interp));
static int 		Win32Pipeline _ANSI_ARGS_((Tcl_Interp *interp,
			    int *pidPtr, int *numPids, int argc, char **argv,
			    Tcl_File inputFile,
			    Tcl_File outputFile,
			    Tcl_File errorFile));
static int 		Win32sPipeline _ANSI_ARGS_((Tcl_Interp *interp,
			    int argc, char **argv, char *IntIn,
			    char *FinalOut, int *pidPtr));

#define RunningNT()	((GetVersion() & 0x80000000) == 0)
#define RunningWin95()	((GetVersion() & 0x800000ff) == 0x80000004)
#define RunningWin32s()	((GetVersion() & 0x800000ff) == 0x80000003)

/*
 * The following defines are the possible values returned by the
 * ApplicationType function.  They identify the various types of
 * applications that run under windows.
 */

#define APPL_NONE	0
#define APPL_DOS	1
#define APPL_WIN3X	2
#define APPL_WIN32	3


/*
 *----------------------------------------------------------------------
 *
 * TclSpawnPipeline --
 *
 *      Calls either Win32Pipeline or Win32sPipeline, depending on which
 *      version of Windows is running.
 *
 * Results:
 *      The return value is 1 on success, 0 on error
 *
 * Side effects:
 *      Processes and pipes are created.
 *
 *----------------------------------------------------------------------
 */

int
TclSpawnPipeline(interp, pidPtr, numPids, argc, argv, inputFile, outputFile,
	errorFile, intIn, finalOut)
    Tcl_Interp *interp;		/* Interpreter in which to process pipeline. */
    int *pidPtr;		/* Array of pids which are created. */
    int *numPids;		/* Number of pids created. */
    int argc;			/* Number of entries in argv. */
    char **argv;		/* Array of strings describing commands in
				 * pipeline plus I/O redirection with <,
				 * <<, >, etc. argv[argc] must be NULL. */
    Tcl_File inputFile;	/* If >=0, gives file id to use as input for
				 * first process in pipeline (specified via <
				 * or <@). */
    Tcl_File outputFile;	/* Writable file id for output from last
				 * command in pipeline (could be file or
				 * pipe). NULL means use stdout. */
    Tcl_File errorFile;	/* Writable file id for error output from all
				 * commands in the pipeline. NULL means use
				 * stderr */
    char *intIn;		/* File name for initial input (for Win32s). */
    char *finalOut;		/* File name for final output (for Win32s). */
{
    if (TclHasPipes()) {
	return(Win32Pipeline(interp, pidPtr, numPids, argc, argv, inputFile,
		outputFile, errorFile));
    }

    return(Win32sPipeline(interp, argc, argv, intIn, finalOut, pidPtr));
}

/*
 *----------------------------------------------------------------------
 *
 * TclpCreateProcess --
 *
 *	Create a child process that has the specified files as its 
 *	standard input, output, and error.  The child process runs
 *	asynchronously and runs with the same environment variables
 *	as the creating process.
 *
 *	The path is searched to find the specified executable.  If an
 *	executable by the given name is not found, automatically tries 
 *	appending ".com", ".exe", and ".bat" to the executable name.
 *
 * Results:
 *	The return value is TCL_ERROR and an error message is left in
 *	interp->result if there was a problem creating the child 
 *	process.  Otherwise, the return value is TCL_OK and *pidPtr is
 *	filled with the process id of the child process.
 * 
 * Side effects:
 *	Processes are created.
 *	
 *----------------------------------------------------------------------
 */

int
TclpCreateProcess(interp, argc, argv, inputFile, outputFile, errorFile, 
	pidPtr)
    Tcl_Interp *interp;		/* Interpreter in which to leave errors that
				 * occurred when creating the child process.
				 * Error messages from the child process 
				 * itself are sent to errorFile. */
    int argc;			/* Number of arguments in following array. */
    char **argv;		/* Array of argument strings.  argv[0] 
				 * contains the name of the executable. */
    Tcl_File inputFile;		/* If non-NULL, gives the file to use as
				 * input for the child process.  If inputFile
				 * file is not readable or is NULL, the child 
				 * will receive no standard input. */
    Tcl_File outputFile;	/* If non-NULL, gives the file that receives  
				 * output from the child process.  If 
				 * outputFile file is not writeable or is 
				 * NULL, output from the child will be 
				 * discarded. */
    Tcl_File errorFile;		/* If non-NULL, gives the file that receives 
				 * errors from the child process.  If 
				 * errorFile file is not writeable or is NULL,
				 * errors from the child will be discarded.
				 * errorFile may be the same as outputFile. */
    int *pidPtr;		/* If this procedure is successful, pidPtr is
				 * filled with the process id of the child 
				 * process. */
{
    int result, len, type, i, searchLen, applType, createFlags, tempFile;
    Tcl_DString cmdLine;
    STARTUPINFO startInfo;
    PROCESS_INFORMATION procInfo;
    SECURITY_ATTRIBUTES secAtts;
    HANDLE hProcess, h, inputHandle, outputHandle, errorHandle;
    char execPath[MAX_PATH], buf[MAX_PATH];
    char *originalName;
    static char extensions[][5] = {".com", ".exe", ".bat"};

    result = TCL_ERROR;
    tempFile = 0;
    Tcl_DStringInit(&cmdLine);

    memset(&startInfo, 0, sizeof(startInfo));
    startInfo.cb	= sizeof(startInfo);
    startInfo.dwFlags   = STARTF_USESTDHANDLES;
    startInfo.hStdInput	= INVALID_HANDLE_VALUE;
    startInfo.hStdOutput= INVALID_HANDLE_VALUE;
    startInfo.hStdError = INVALID_HANDLE_VALUE;

    inputHandle		= INVALID_HANDLE_VALUE;
    outputHandle	= INVALID_HANDLE_VALUE;
    errorHandle		= INVALID_HANDLE_VALUE;

    secAtts.nLength = sizeof(SECURITY_ATTRIBUTES);
    secAtts.lpSecurityDescriptor = NULL;
    secAtts.bInheritHandle = TRUE;

    hProcess = GetCurrentProcess();

    /* 
     * Find the executable.
     *
     * Using the raw SearchPath() command doesn't do quite what is 
     * necessary.  If the name of the executable contains a '.' 
     * character, it will not try appending the specified extension when
     * searching (in other words, SearchPath will not find the program 
     * "a.b.exe" if the arguments specified "a.b" and ".exe").   
     *
     * So, first look for the file as it is named.  Then manually append 
     * the extensions, looking for a match.  
     *
     * If it can't find any executable that matched, then it will cause an
     * (expected) error about not being able to find the program,
     * in CreateProcess below.
     */

    lstrcpyn(buf, argv[0], sizeof(buf) - 5);
    len = strlen(buf);

    for (i = -1; i < (int) (sizeof(extensions) / sizeof(extensions[0])); i++) {
	if (i >= 0) {
	    strcpy(buf + len, extensions[i]);
	}
	
	searchLen = SearchPath(NULL, buf, NULL, sizeof(execPath), 
		execPath, NULL);
	if ((searchLen != 0) && (searchLen < sizeof(execPath)) 
		&& ((GetFileAttributes(execPath) 
			& FILE_ATTRIBUTE_DIRECTORY) == 0)) {
	    break;
	}
    }

    originalName = argv[0];
    argv[0] = execPath;
    applType = ApplicationType(execPath);

    if ((applType == APPL_DOS) || (applType == APPL_WIN3X)) {
	/* 
	 * Replace long path name of executable with short path name for 
	 * 16-bit applications.  Otherwise the application may not be able
	 * to correctly parse its own command line to separate off the 
	 * application name from the arguments.
	 */

	GetShortPathName(execPath, execPath, sizeof(execPath));
    }


    /*
     * We have to check the type of each file, since we cannot duplicate 
     * some file types.  
     */

    if (inputFile != NULL) {
	h = (HANDLE) Tcl_GetFileInfo(inputFile, &type);
	if ((type >= TCL_WIN_PIPE) && (type <= TCL_WIN_CONSOLE)) {
	    inputHandle = h;
	}
    }
    if (outputFile != NULL) {
	h = (HANDLE) Tcl_GetFileInfo(outputFile, &type);
	if ((type >= TCL_WIN_PIPE) && (type <= TCL_WIN_CONSOLE)) {
	    outputHandle = h;
	}
    }
    if (errorFile != NULL) {
	h = (HANDLE) Tcl_GetFileInfo(errorFile, &type);
	if ((type >= TCL_WIN_PIPE) && (type <= TCL_WIN_CONSOLE)) {
	    errorHandle = h;
	}
    }

    /*
     * Duplicate all the handles which will be passed off as stdin, stdout
     * and stderr of the child process. The duplicate handles are set to
     * be inheritable, so the child process can use them.
     */

    if (inputHandle == INVALID_HANDLE_VALUE) {
	/* 
	 * If handle was not set, stdin should return immediate EOF.
	 * Under Windows95, some applications (both 16 and 32 bit!) 
	 * cannot detect EOF when redirecting from NUL.  Fortunately, 
	 * all apps can (now) detect a closed pipe as EOF.
	 */

	HANDLE hReadNULL, hWriteNULL;
	CreatePipe(&hReadNULL, &hWriteNULL, &secAtts, 0);
	CloseHandle(hWriteNULL);
	startInfo.hStdInput = hReadNULL;
    } else {
	DuplicateHandle(hProcess, inputHandle, hProcess, &startInfo.hStdInput,
		0, TRUE, DUPLICATE_SAME_ACCESS);
    }
    if (startInfo.hStdInput == INVALID_HANDLE_VALUE) {
	TclWinConvertError(GetLastError());
	Tcl_AppendResult(interp, "couldn't duplicate input handle: ",
		Tcl_PosixError(interp), (char *) NULL);
        goto end;
    }

    if (outputHandle == INVALID_HANDLE_VALUE) {
	/*
	 * If handle was not set, output should be sent to an infinitely 
	 * deep sink.  Under Windows 95, some 16 bit applications cannot 
	 * have stdout redirected to NUL when they are outputting large
	 * amounts of data.  Why????  Redirect stdout in that case 
	 * to a temp file that eventually gets deleted.
	 */

	if ((applType == APPL_DOS) && RunningWin95()) {
	    buf[0] = '\0';
	    GetTempPath(sizeof(buf), buf);
	    GetTempFileName(buf, "tcl", 0, buf);
	    startInfo.hStdOutput = CreateFile(buf, 
		    GENERIC_READ | GENERIC_WRITE, 0, &secAtts, OPEN_ALWAYS, 
		    FILE_ATTRIBUTE_TEMPORARY | FILE_FLAG_DELETE_ON_CLOSE,
		    NULL);
	    tempFile = 1;
	} else {
	    startInfo.hStdOutput = CreateFile("NUL:", GENERIC_WRITE, 0,
		    &secAtts, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	}
    } else {
	DuplicateHandle(hProcess, outputHandle, hProcess, &startInfo.hStdOutput, 
		0, TRUE, DUPLICATE_SAME_ACCESS);
    }
    if (startInfo.hStdOutput == INVALID_HANDLE_VALUE) {
	TclWinConvertError(GetLastError());
	Tcl_AppendResult(interp, "couldn't duplicate output handle: ",
		Tcl_PosixError(interp), (char *) NULL);
	goto end;
    }

    if (errorHandle == INVALID_HANDLE_VALUE) {
	/*
	 * If handle was not set, errors should be sent to an infinitely 
	 * deep sink.  
	 */

	startInfo.hStdError = CreateFile("NUL:", GENERIC_WRITE, 0,
		&secAtts, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    } else {
	DuplicateHandle(hProcess, errorHandle, hProcess, &startInfo.hStdError, 
		0, TRUE, DUPLICATE_SAME_ACCESS);
    } 
    if (startInfo.hStdError == INVALID_HANDLE_VALUE) {
	TclWinConvertError(GetLastError());
	Tcl_AppendResult(interp, "couldn't duplicate error handle: ",
		Tcl_PosixError(interp), (char *) NULL);
	goto end;
    }

    /* 
     * If we do not have a console window, then we must run DOS and
     * WIN32 console mode applications as detached processes. This tells
     * the loader that the child application should not inherit the
     * console, and that it should not create a new console window for
     * the child application.  The child application should get its stdio 
     * from the redirection handles provided by this application, and run
     * in the background.
     *
     * If we are starting a GUI process, they don't automatically get a 
     * console, so it doesn't matter if they are started as foreground or
     * detached processes.  The GUI window will still pop up to the
     * foreground.
     */

    if (RunningNT()) {
	if (HasConsole()) {
	    createFlags = 0;
	} else if (applType == APPL_DOS) {
	    /*
	     * Under NT, 16-bit DOS applications will not run unless they
	     * can be attached to a console.  If we are running without a
	     * console, run the 16-bit program as an normal process inside
	     * of a hidden console application, and then run that hidden
	     * console as a detached process.  Hey, it works.
	     */

	    Tcl_DStringAppend(&cmdLine, "cmd.exe /c ", -1);
	    startInfo.wShowWindow = SW_HIDE;
	    startInfo.dwFlags |= STARTF_USESHOWWINDOW;
	    createFlags = CREATE_NEW_CONSOLE;
	} else {
	    createFlags = DETACHED_PROCESS;
	}
    } else {	/* Running95() */

	if (HasConsole()) {
	    createFlags = 0;
	} else {
	    createFlags = DETACHED_PROCESS;
	}
	
	if (applType == APPL_DOS) {
	    /*
	     * Under Windows 95, 16-bit DOS applications do not work well 
	     * with pipes:
	     *
	     * 1. EOF on a pipe between a detached 16-bit DOS application 
	     * and another application is not seen at the other
	     * end of the pipe, so the listening process blocks forever on 
	     * reads.  This inablity to detect EOF happens when either a 
	     * 16-bit app or the 32-bit app is the listener.  
	     *
	     * 2. If a 16-bit DOS application (detached or not) blocks when 
	     * writing to a pipe, it will never wake up again, and it
	     * eventually brings the whole system down around it.
	     *
	     * The 16-bit application is run as a normal process inside
	     * of a hidden helper console app, and this helper may be run
	     * as a detached process.  If any of the stdio handles is
	     * a pipe, the helper application accumulates information 
	     * into temp files and forwards it to or from the DOS 
	     * application as appropriate.  This means that DOS apps 
	     * must receive EOF from a stdin pipe before they will actually
	     * begin, and must finish generating stdout or stderr before 
	     * the data will be sent to the next stage of the pipe.
	     *
	     * The helper app should be located in the same directory as
	     * the tcl dll.
	     */

	    Tcl_DStringAppend(&cmdLine, "stub16.exe \"", -1);
	    if (tempFile) { 
		Tcl_DStringAppend(&cmdLine, buf, -1);
	    }
	    Tcl_DStringAppend(&cmdLine, "\" ", 2);
	    if (createFlags != 0) {
		startInfo.wShowWindow = SW_HIDE;
		startInfo.dwFlags |= STARTF_USESHOWWINDOW;
		createFlags = CREATE_NEW_CONSOLE;
	    }
	}
    }

    /*
     * cmdLine gets the full command line used to invoke the executable,
     * including the name of the executable itself.  The command line
     * arguments in argv[] are stored in cmdLine separated by spaces. 
     * Special characters in individual arguments from argv[] must be 
     * quoted when being stored in cmdLine.
     *
     * When calling any application, bear in mind that arguments that 
     * specify a path name are not converted.  If an argument contains 
     * forward slashes as path separators, it may or may not be 
     * recognized as a path name, depending on the program.  In general,
     * most applications accept forward slashes only as option 
     * delimiters and backslashes only as paths.
     *
     * Additionally, when calling a 16-bit dos or windows application, 
     * all path names must use the short, cryptic, path format (e.g., 
     * using ab~1.def instead of "a b.default").  
     */

    for (i = 0; i < argc; i++) {
	char *start, *special;
	int quote;

	if (i > 0) {
	    Tcl_DStringAppend(&cmdLine, " ", 1);	
	}

	quote = 0;
	if ((strchr(argv[i], ' ') != NULL) || (strchr(argv[i], '\t') != NULL)) {
	    quote = 1;
	    Tcl_DStringAppend(&cmdLine, "\"", 1);
	}
	
	start = argv[i];	    
	for (special = argv[i]; ; ) {
	    if ((*special == '\\') && 
		    (special[1] == '\\' || special[1] == '"')) {
		Tcl_DStringAppend(&cmdLine, start, special - start);
		start = special;
		while (1) {
		    special++;
		    if (*special == '"') {
			/* 
			 * N backslashes followed a quote -> insert 
			 * N * 2 + 1 backslashes then a quote.
			 */

			Tcl_DStringAppend(&cmdLine, start, special - start);
			break;
		    }
		    if (*special != '\\') {
			break;
		    }
		}
		Tcl_DStringAppend(&cmdLine, start, special - start);
		start = special;
	    }
	    if (*special == '"') {
		Tcl_DStringAppend(&cmdLine, start, special - start);
		Tcl_DStringAppend(&cmdLine, "\\\"", 2);
		start = special + 1;
	    }
	    if (*special == '\0') {
		break;
	    }
	    special++;
	}
	Tcl_DStringAppend(&cmdLine, start, special - start);
	if (quote) {
	    Tcl_DStringAppend(&cmdLine, "\"", 1);
	}
    }

    if (!CreateProcess(NULL, Tcl_DStringValue(&cmdLine), NULL, NULL, TRUE, 
	    createFlags, NULL, NULL, &startInfo, &procInfo)) {
	TclWinConvertError(GetLastError());
	Tcl_AppendResult(interp, "couldn't execute \"", originalName,
		"\": ", Tcl_PosixError(interp), (char *) NULL);
	goto end;
    }

    if ((applType == APPL_DOS)) {
	/* 
	 * When running a DOS app, it seems that unless we put a small
	 * pause here, we end up closing the stdio handles in the next few 
	 * lines before the DOS app laucher has had a chance to inherit 
	 * those handles.  This causes the DOS app to get an invalid handle
	 * and send its output or error to a new console window.
	 * This is not a problem with Windows 3.X apps (because they have
	 * no stdio) or Win32 apps.
	 *
	 * Incidentaly, WaitForSingleObject(hProcess, 1) didn't seem to
	 * give the DOS app enough time to get the handles, so it wasn't
	 * just a context switch that was necessary.
	 */

	WaitForSingleObject(hProcess, 50);
    }
    *pidPtr = (int) procInfo.hProcess;
    CloseHandle(procInfo.hThread);
    result = TCL_OK;

end:
    Tcl_DStringFree(&cmdLine);
    if (startInfo.hStdInput != INVALID_HANDLE_VALUE) {
        CloseHandle(startInfo.hStdInput);
    }
    if (startInfo.hStdOutput != INVALID_HANDLE_VALUE) {
        CloseHandle(startInfo.hStdOutput);
    }
    if (startInfo.hStdError != INVALID_HANDLE_VALUE) {
	CloseHandle(startInfo.hStdError);
    }
    return result;
}
/*
 *----------------------------------------------------------------------
 *
 * Win32Pipeline --
 *
 *      Given an argc/argv array, instantiate a pipeline of processes
 *      as described by the argv.
 *
 * Results:
 *      The return value is 1 on success, 0 on error
 *
 * Side effects:
 *      Processes and pipes are created.
 *
 *----------------------------------------------------------------------
 */

static int
Win32Pipeline(interp, pidPtr, numPids, argc, argv, inputFile, outputFile,
	errorFile)
    Tcl_Interp *interp;		/* Interpreter in which to process pipeline. */
    int *pidPtr;		/* Array of pids which are created. */
    int *numPids;		/* Number of pids created. */
    int argc;			/* Number of entries in argv. */
    char **argv;		/* Array of strings describing commands in
				 * pipeline plus I/O redirection with <,
				 * <<, >, etc. argv[argc] must be NULL. */
    Tcl_File inputFile;	/* If >=0, gives file id to use as input for
				 * first process in pipeline (specified via <
				 * or <@). */
    Tcl_File outputFile;	/* Writable file id for output from last
				 * command in pipeline (could be file or
				 * pipe). NULL means use stdout. */
    Tcl_File errorFile;	/* Writable file id for error output from all
				 * commands in the pipeline. NULL means use
				 * stderr */
{
    Tcl_Channel channel;
    int firstArg, lastArg, pid;
    char *exec;
    Tcl_DString name;
    int joinThisError;
    Tcl_File pipeIn, curOutFile, curInFile, curErrFile;
    Tcl_File stdInFile, stdOutFile, stdErrFile;
    
    /*
     * Get the current stdio files, to be used if any of the provided 
     * IO files are NULL.   We need stdInFile, stdOutFile, and stdErrFile
     * further below when closing any temporary files, to make sure that we 
     * don't close these handles.
     */

    stdInFile = NULL;
    stdOutFile = NULL;
    stdErrFile = NULL;
    
    channel = Tcl_GetStdChannel(TCL_STDIN);
    if (channel) {
	stdInFile = Tcl_GetChannelFile(channel, TCL_READABLE);
    }
    channel = Tcl_GetStdChannel(TCL_STDOUT);
    if (channel) {
	stdOutFile = Tcl_GetChannelFile(channel, TCL_WRITABLE);
    }
    channel = Tcl_GetStdChannel(TCL_STDERR);
    if (channel) {
	stdErrFile = Tcl_GetChannelFile(channel, TCL_WRITABLE);
    }
    
    if (inputFile == NULL) {
	inputFile = stdInFile;
    }

    if (outputFile == NULL) {
	outputFile = stdOutFile;
    }

    if (errorFile == NULL) {
	errorFile = stdErrFile;
    }

    pipeIn = NULL;
    curOutFile = NULL;
    curInFile = inputFile;

    Tcl_DStringInit(&name);

    for (firstArg = 0; firstArg < argc; firstArg = lastArg+1) { 

	/*
	 * Convert the program name into native form. 
	 */

	exec = Tcl_TranslateFileName(interp, argv[firstArg], &name);
	if (exec == NULL) {
	    goto error;
	}

	/*
	 * Find the end of the current segment of the pipeline.
	 */

	joinThisError = 0;
	for (lastArg = firstArg; lastArg < argc; lastArg++) {
	    if (argv[lastArg][0] == '|') { 
		if (argv[lastArg][1] == 0) { 
		    break;
		}
		if ((argv[lastArg][1] == '&') && (argv[lastArg][2] == 0)) {
		    joinThisError = 1;
		    break;
		}
	    }
	}

	/*
	 * If this is the last segment, use the specified outputFile.
	 * Otherwise create an intermediate pipe.  pipeIn will become the
	 * curInFile for the next segment of the pipe.
	 */

	if (lastArg == argc) { 
	    curOutFile = outputFile;
	} else {
	    if (TclCreatePipe(&pipeIn, &curOutFile) == 0) {
		Tcl_AppendResult(interp, "couldn't create pipe: ",
			Tcl_PosixError(interp), (char *) NULL);
		goto error;
	    }
	}

	if (joinThisError != 0) {
	    curErrFile = curOutFile;
	} else {
	    curErrFile = errorFile;
	}

	if (TclpCreateProcess(interp, lastArg - firstArg, argv + firstArg,
		curInFile, curOutFile, curErrFile, &pid) != TCL_OK) {
	    goto error;
	}

	/*
	 * Add the child process to the list of those to be reaped.  Also,
	 * close the thread handle, since it won't be used.
	 */

	pidPtr[*numPids] = pid;
	(*numPids)++;

	/*
	 * Close off our copies of file descriptors that were set up for
	 * this child, then set up the input for the next child.
	 */

	if ((curInFile != NULL) && (curInFile != inputFile) 
		&& (curInFile != stdInFile)) {
	    TclCloseFile(curInFile);
	}
	curInFile = pipeIn;
	pipeIn = NULL;

	if ((curOutFile != NULL) && (curOutFile != outputFile)
		&& (curOutFile != stdOutFile)) {
	    TclCloseFile(curOutFile);
	}
	curOutFile = NULL;
    }
    return 1;

    /*
     * An error occured, so we need to clean up any open pipes.
     */

error:
    Tcl_DStringFree(&name);
    if (pipeIn) {
	TclCloseFile(pipeIn);
    }
    if ((curOutFile != NULL) && (curOutFile != outputFile)
		&& (curOutFile != stdOutFile)) {
	TclCloseFile(curOutFile);
    }
    if ((curInFile != NULL) && (curInFile != inputFile)
		&& (curInFile != stdInFile)) {
	TclCloseFile(curInFile);
    }
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * HasConsole --
 *
 *	Determines whether the current application is attached to a
 *	console.
 *
 * Results:
 *	Returns TRUE if this application has a console, else FALSE.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static BOOL
HasConsole()
{
    HANDLE handle = CreateFile("CONOUT$", GENERIC_WRITE, FILE_SHARE_WRITE,
	    NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

    if (handle != INVALID_HANDLE_VALUE) {
        CloseHandle(handle);
	return TRUE;
    } else {
        return FALSE;
    }
}

/*
 *--------------------------------------------------------------------
 *
 * ApplicationType --
 *
 *	Used to determine if the specified name refers to a DOS,
 *	Windows 3.X, or Win32 program.  Used to determine how to invoke 
 *	a program, or if it can even be invoked.
 *
 *	It is possible to almost positively identify DOS and Windows 
 *	applications that contain the appropriate magic numbers.  However, 
 *	DOS .com files do not seem to contain a magic number; if the program 
 *	name ends with .com and could not be identified as a Windows .com
 *	file, it will be assumed to be a DOS application, even if it was
 *	just random data.  If the program name does not end with .com, no 
 *	such assumption is made.
 *
 *	The Win32 procedure GetBinaryType incorrectly identifies any 
 *	junk file that ends with .exe as a dos executable and some 
 *	executables that don't end with .exe as not executable.  Plus it 
 *	doesn't exist under win95, so I won't feel bad about reimplementing
 *	functionality.
 *
 * Results:
 *	The return value is one of APPL_DOS, APPL_WIN3X, or APPL_WIN32
 *	if the filename referred to the corresponding application type,
 *	or APPL_NONE if the filename was none of the above.  .BAT files
 *	are identified as APPL_NONE, but that's OK because nothing special
 *	is done for WIN32 functions either.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
ApplicationType(fileName)
    const char *fileName;
{
    HANDLE hFile;
    unsigned long header;
    DWORD read;
    char buf[0x40];
    
    hFile = CreateFile(fileName, GENERIC_READ, FILE_SHARE_READ, NULL, 
	    OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile == INVALID_HANDLE_VALUE) {
	return FALSE;
    }
    
    if ((ReadFile(hFile, (void *) buf, 0x40, &read, NULL) == FALSE) 
	    || (read != 0x40)) {
	CloseHandle(hFile);
	return FALSE;
    }

    if ((buf[0] != 'M') || (buf[1] != 'Z')) {
	/* 
	 * Doesn't have the magic number for relocatable executables.  If 
	 * filename ends with .com it's probably a DOS application anyhow.
	 */

	CloseHandle(hFile);
	fileName = strrchr(fileName, '.');
	if ((fileName != NULL) && (strcmpi(fileName, ".com") == 0)) {
	    return APPL_DOS;
	}
	return APPL_NONE;
    }

    if (*((short *) (buf + 0x18)) != 0x0040) {
	/* 
	 * All Windows 3.X and Win32 and some DOS programs have this magic
	 * number here.  If it doesn't, assume that since it already had the 
	 * other magic number it was a DOS application.
	 */

	CloseHandle(hFile);
	return APPL_DOS;
    }

    /* 
     * The DWORD at 0x3C points to yet another magic number.
     */

    header = *((long *) (buf + 0x3C));
    if (SetFilePointer(hFile, header, NULL, FILE_BEGIN) != header) {
	CloseHandle(hFile);
	return APPL_NONE;
    }
    ReadFile(hFile, (void *) buf, 2, &read, NULL);
    CloseHandle(hFile);

    if ((buf[0] == 'L') && (buf[1] == 'E')) {
	return APPL_DOS;
    }
    if ((buf[0] == 'N') && (buf[1] == 'E')) {
	return APPL_WIN3X;
    }
    return APPL_WIN32;
}

/*
 *----------------------------------------------------------------------
 *
 * TclHasPipes --
 *
 *      Checks to see if it's running under Win32s or Win32.
 *
 * Results:
 *      Returns 1 if running under Win32, 0 otherwise.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
TclHasPipes(void)
{
    OSVERSIONINFO info;
    info.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
    GetVersionEx(&info);
    return((info.dwPlatformId == VER_PLATFORM_WIN32s) ? 0 : 1);
}

/*
 *----------------------------------------------------------------------
 *
 * TclCreatePipe --
 *
 *      Creates an anonymous pipe.
 *
 * Results:
 *      Returns 1 on success, 0 on failure. 
 *
 * Side effects:
 *      Creates a pipe.
 *
 *----------------------------------------------------------------------
 */

int
TclCreatePipe(readPipe, writePipe)
    Tcl_File *readPipe;	/* Location to store file handle for
				 * read side of pipe. */
    Tcl_File *writePipe;	/* Location to store file handle for
				 * write side of pipe. */
{
    HANDLE readHandle, writeHandle;
    SECURITY_ATTRIBUTES sec;

    sec.nLength = sizeof(SECURITY_ATTRIBUTES);
    sec.lpSecurityDescriptor = NULL;
    sec.bInheritHandle = FALSE;

    if (!CreatePipe(&readHandle, &writeHandle, &sec, 0)) {
	TclWinConvertError(GetLastError());
	return 0;
    }

    *readPipe = Tcl_GetFile((ClientData)readHandle, TCL_WIN_FILE);
    *writePipe = Tcl_GetFile((ClientData)writeHandle, TCL_WIN_FILE);
    return 1;
}

/*
 *----------------------------------------------------------------------
 *
 * Win32sPipeline --
 *
 *      Given an argc/argv array, instantiate a pipeline of processes
 *      as described by the argv. This is the Win32s version, which works
 *      by creating a batch file to run the pipeline.
 *
 * Results:
 *      The return value is 1 on success, 0 on error
 *
 * Side effects:
 *      A batch file describing the pipeline is executed.
 *
 *----------------------------------------------------------------------
 */

static int
Win32sPipeline(interp, argc, argv, intIn, finalOut, pidPtr)
    Tcl_Interp *interp;		/* Interpreter in which to process pipeline. */
    int argc;			/* Number of entries in argv. */
    char **argv;		/* Array of strings describing commands in
				 * pipeline plus I/O redirection with <,
				 * <<, >, etc. argv[argc] must be NULL. */
    char *intIn;		/* File name for initial input (for Win32s). */
    char *finalOut;		/* File name for final output (for Win32s). */
    int *pidPtr;		/* Array of pids which are created. */
{
    char batchFile[20];		/* Name of batch file */
    int batchHandle;		/* Handle to the batch file */ 
    int outputHandle;		/* Handle to final output file */
    char string[100];		/* Temporary string */
    int numBytes;
    int firstArg, lastArg;	/* Indices of first and last arguments
				 * in current command. */
    Tcl_DString execProc;	/* Dynamic string used to construct
				 * intermediate commands */
    char *ptr;			/* Temporary pointer - used to build
				 * execProc strings */
    char intOut[L_tmpnam];	/* Name of temporary file */
    int result = 1;		/* Return value of this function. */
    int dirLength, offset;
    
    Tcl_DStringInit(&execProc);

    /*
     * Create batch file
     */

    tmpnam(batchFile);
    ptr = strchr(batchFile, '.');
    strcpy(ptr, ".bat");
    batchHandle = open(batchFile, O_CREAT | O_WRONLY | O_TEXT, 0600);
    if (batchHandle < 0) {
	Tcl_AppendResult(interp,
		"couldn't create batch file \"", batchFile,
		"\" for command: ", Tcl_PosixError(interp),
		(char *) NULL);
	result = 0;
	goto cleanup;
    }

    /*
     * Cycle through the pipeline, generating commands to place in batch file.
     */ 

    for (firstArg = 0; firstArg < argc; firstArg = lastArg+1) { 
	Tcl_DStringSetLength(&execProc, 0);
	for (lastArg = firstArg; lastArg < argc; lastArg++) {
	    if (argv[lastArg][0] == '|') {
		if (argv[lastArg][1] == 0) {
		    break;
		}
		if ((argv[lastArg][1] == '&') && (argv[lastArg][2] == 0)) {
		    break;
		}
	    } else {
		Tcl_DStringAppend(&execProc, argv[lastArg], -1);
		Tcl_DStringAppend(&execProc, " ", -1);
	    }
	}

	/*
	 * Check if this pipeline command is a Windows application.
	 */

	if (IsWinApp(argv[firstArg], interp)) {
	    if ((firstArg != 0) && (lastArg != argc)) {
		Tcl_AppendResult(interp,
			"Can't pipe input in or out of Windows app", finalOut,
			(char *) NULL);
		result = 0;
		close(batchHandle);
		goto cleanup;
	    }
	}

	/*
	 * Set input redirection for the current command in the pipeline.
	 */

	if (intIn[0]) {
	    Tcl_DStringAppend(&execProc, " < ", -1);
	    Tcl_DStringAppend(&execProc, intIn, -1);
	}

	/*
	 * Set output redirection for the current command in the pipeline.
	 */

	tmpnam(intOut);
	Tcl_DStringAppend(&execProc, " > ", -1);
	if (lastArg == argc) {
	    Tcl_DStringAppend(&execProc, finalOut, -1);
	} else {
	    Tcl_DStringAppend(&execProc, intOut, -1);
	}
	Tcl_DStringAppend(&execProc, "\r\n", -1);
	
	if (intIn[0]) {
	    Tcl_DStringAppend(&execProc, "del ", -1);
	    Tcl_DStringAppend(&execProc, intIn, -1);
	    Tcl_DStringAppend(&execProc, "\r\n", -1);
	}

	write(batchHandle, Tcl_DStringValue(&execProc),
		Tcl_DStringLength(&execProc));

	strcpy(intIn, intOut);
    }
    close(batchHandle);

    /*
     * Set up the command to execute the batch file
     */

    dirLength = GetCurrentDirectory(0, NULL);
    Tcl_DStringSetLength(&execProc, 0);
    Tcl_DStringAppend(&execProc, "command.com /C ", -1);
    offset = Tcl_DStringLength(&execProc);
    Tcl_DStringSetLength(&execProc, (dirLength - 1) + offset);
    GetCurrentDirectory(dirLength, Tcl_DStringValue(&execProc) + offset);
    Tcl_DStringAppend(&execProc, "\\", -1);
    Tcl_DStringAppend(&execProc, batchFile, -1);

    /*
     * Do a synchronous spawn of the batch file
     */

    TclSynchSpawn(Tcl_DStringValue(&execProc), SW_SHOWNORMAL);

    /*
     * Read the output from the command pipeline
     */

    outputHandle = open(finalOut, O_RDONLY | O_TEXT, 0600);
    if (outputHandle < 0) { 
	Tcl_AppendResult(interp,
		"couldn't read output file \"", finalOut,
		"\" for command: ", Tcl_PosixError(interp),
		(char *) NULL);
	result = 0;
	goto cleanup;
    }
    
    do {
	numBytes = read(outputHandle, string, 100);
	string[numBytes] = '\0';
	Tcl_AppendResult(interp, string, (char *)NULL);
    } while(numBytes > 0);
    close(outputHandle);

  cleanup:
    unlink(batchFile);
    unlink(finalOut);
    
    Tcl_DStringFree(&execProc);
    return(result);
}

/*
 *----------------------------------------------------------------------
 *
 * IsWinApp --
 *
 *      Given a filename, checks to see if that file is a Windows executable.
 *
 * Results:
 *      TRUE if it is a Windows executable, FALSE otherwise.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static BOOL
IsWinApp(filename, interp)
    char *filename;                /* Name of file to check */
    Tcl_Interp *interp;            /* Interpreter in which pipeline is being
				    * processed. */ 
{
    int fileId;
    char buffer[MAX_PATH+1], *dummy;

   /*
    * If the word value at offset 18h of the file is 40h or greater,
    * the word value at 3Ch is an offset to a Windows header. Sooo.... we
    * must read the file header and check that offset to determine whether
    * "filename" is a Windows app.
    */

    if (!SearchPath(NULL, filename, ".exe", MAX_PATH, buffer, &dummy)) {
	return FALSE;
    }
	
    fileId = open(buffer, O_RDONLY, 0600);
    if (fileId < 0) {
	Tcl_AppendResult(interp,
		"couldn't open file \"", filename,
		"\" for command: ", Tcl_PosixError(interp),
		(char *) NULL);
    } else {
	read(fileId, buffer, MAX_PATH);
	close(fileId);
	if (buffer[24] >= 64) {
	    return(TRUE);
	}
    }
    return(FALSE);
}
